from numpy import NaN
from pandas import DataFrame, Series, Timestamp

from ._identifiers import IDENTIFIERS
from .._type_converter import infer_type, to_bool, str_to_datetime, datetime_to_str, type2str
from ..helpers import BaseSectionObject, split_line_with_quotes
from ..section_labels import *


class RainGage(BaseSectionObject):
    """
    Section: [**RAINGAGES**]

    Purpose:
        Identifies each rain gage that provides rainfall data for the study area.

    Formats:
        ::

            Name Form Intvl SCF TIMESERIES Tseries
            Name Form Intvl SCF FILE       Fname   Sta Units

    Format-PCSWMM:
        ``Name Format Interval SCF Source``


    Attributes:
        Name (str): name assigned to rain gage.
        Format (str): form of recorded rainfall, either INTENSITY, VOLUME or CUMULATIVE.
        Interval (str, Timedelta): time interval between gage readings in decimal hours or hours:minutes format
                                    (e.g., 0:15 for 15-minute readings). ``Intvl``
        SCF (float): snow catch deficiency correction factor (use 1.0 for no adjustment).
        Source (str): one of ``'TIMESERIES'`` ``'FILE'``
        Timeseries (str): name of time series in [TIMESERIES] section with rainfall data. ``Tseries``
        Filename (str): name of external file with rainfall data.
                        Rainfall files are discussed in Section 11.3 Rainfall Files. ``Fname``
        Station (str): name of recording station used in the rain file. ``Sta``
        Units (str): rain depth units used in the rain file, either IN (inches) or MM (millimeters).
    """
    _identifier = IDENTIFIERS.Name
    _section_label = RAINGAGES

    class FORMATS:
        INTENSITY = 'INTENSITY'
        VOLUME = 'VOLUME'
        CUMULATIVE = 'CUMULATIVE'

    class SOURCES:
        TIMESERIES = 'TIMESERIES'
        FILE = 'FILE'

    class UNITS:
        IN = 'IN'
        MM = 'MM'

    def __init__(self, Name, Format, Interval, SCF, Source, *args,
                 Timeseries=NaN,
                 Filename=NaN, Station=NaN, Units=NaN):
        """
        Object for section RAINGAGES

        Args:
            Name (str): name assigned to rain gage.
            Format (str): form of recorded rainfall, either INTENSITY, VOLUME or CUMULATIVE.
            Interval (str, Timedelta): time interval between gage readings in decimal hours or hours:minutes format
                                        (e.g., 0:15 for 15-minute readings). ``Intvl``
            SCF (float): snow catch deficiency correction factor (use 1.0 for no adjustment).
            Source (str): one of ``'TIMESERIES'`` ``'FILE'``
            *args: for automatic inp file reading
            Timeseries (:obj:`str`, optional): name of time series in [TIMESERIES] section with rainfall data. ``Tseries``
            Filename (str): name of external file with rainfall data.
                            Rainfall files are discussed in Section 11.3 Rainfall Files. ``Fname``
            Station (str): name of recording station used in the rain file. ``Sta``
            Units (str): rain depth units used in the rain file, either IN (inches) or MM (millimeters).
        """
        self.Name = str(Name)
        self.Format = Format
        self.Interval = Interval
        self.SCF = float(SCF)
        self.Source = Source

        self.Timeseries = Timeseries
        self.Filename = Filename
        self.Station = Station
        self.Units = Units

        if args:
            if Source == RainGage.SOURCES.TIMESERIES:
                self.Timeseries = args[0]
                if len(args) != 1:
                    pass

            elif Source == RainGage.SOURCES.FILE:
                self.Filename = args[0]
                self.Station = args[1]
                self.Units = args[2]

            else:
                raise NotImplementedError()


class Symbol(BaseSectionObject):
    """
    Section: [**SYMBOLS**]

    Purpose:
        Assigns X,Y coordinates to rain gage symbols.

    Format:
        ::

            Gage Xcoord Ycoord

    Args:
        Gage (str): name of gage.
        x (float): horizontal coordinate relative to origin in lower left of map. ``Xcoord``
        y (float): vertical coordinate relative to origin in lower left of map. ``Ycoord``

    Attributes:
        Gage (str): name of gage.
        x (float): horizontal coordinate relative to origin in lower left of map. ``Xcoord``
        y (float): vertical coordinate relative to origin in lower left of map. ``Ycoord``
    """
    _identifier = IDENTIFIERS.Gage
    _section_label = SYMBOLS

    def __init__(self, Gage, x, y):
        self.Gage = str(Gage)
        self.x = float(x)
        self.y = float(y)


class Pattern(BaseSectionObject):
    """
    Section: [**PATTERNS**]

    Purpose:
        Specifies time pattern of dry weather flow or quality in the form of adjustment factors
        applied as multipliers to baseline values.


    Format:
        ::

            Name MONTHLY Factor1 Factor2 ... Factor12
            Name DAILY Factor1  Factor2  ...  Factor7
            Name HOURLY Factor1  Factor2  ...  Factor24
            Name WEEKEND Factor1  Factor2  ...  Factor24

    Remarks:
        The MONTHLY format is used to set monthly pattern factors for dry weather flow constituents.

        The DAILY format is used to set dry weather pattern factors for each day of the week, where Sunday is day 1.

        The HOURLY format is used to set dry weather factors for each hour of the day starting from midnight.
        If these factors are different for weekend days than for weekday days then the WEEKEND format can be used
        to specify hourly adjustment factors just for weekends.

        More than one line can be used to enter a pattern’s factors by repeating the pattern’s name
        (but not the pattern type) at the beginning of each additional line.

        The pattern factors are applied as multipliers to any baseline dry weather flows or quality
        concentrations supplied in the [DWF] section.

    Examples:
        ::

            ; Day of week adjustment factors
            D1 DAILY 0.5 1.0 1.0 1.0 1.0 1.0 0.5
            D2 DAILY 0.8 0.9 1.0 1.1 1.0 0.9 0.8

            ; Hourly adjustment factors
            H1 HOURLY 0.5 0.6 0.7 0.8 0.8 0.9
            H1        1.1 1.2 1.3 1.5 1.1 1.0
            H1        0.9 0.8 0.7 0.6 0.5 0.5
            H1        0.5 0.5 0.5 0.5 0.5 0.5

    Args:
        Name (str): name used to identify the pattern.
        Type (str): one of ``MONTHLY``, ``DAILY``, ``HOURLY``, ``WEEKEND``
        Factors (list): multiplier values.
        *factors: for automatic inp file reading

    Attributes:
        Name (str): name used to identify the pattern.
        Type (str): one of ``MONTHLY``, ``DAILY``, ``HOURLY``, ``WEEKEND``
        Factors (list): multiplier values.
    """
    _identifier = IDENTIFIERS.Name
    _section_label = PATTERNS

    class TYPES:
        MONTHLY = 'MONTHLY'
        DAILY = 'DAILY'
        HOURLY = 'HOURLY'
        WEEKEND = 'WEEKEND'

    def __init__(self, Name, Type, *factors, Factors=None):
        self.Name = str(Name)
        self.Type = Type
        if Factors is not None:
            self.Factors = Factors
        elif isinstance(factors[0], (list, tuple)):
            self.Factors = factors[0]
        else:
            self.Factors = list(float(f) for f in factors)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        args = list()
        for line in multi_line_args:
            if line[1] in [cls.TYPES.MONTHLY, cls.TYPES.DAILY,
                           cls.TYPES.HOURLY, cls.TYPES.WEEKEND]:
                if args:
                    yield cls(*args)
                args = line
            else:
                args += line[1:]
        # last
        if args:
            yield cls(*args)

    def to_inp_line(self):
        if self.Type in (self.TYPES.MONTHLY, self.TYPES.HOURLY, self.TYPES.WEEKEND):
            s = ''
            import numpy as np

            l = len(self.Type)
            first = True
            for a in np.array_split(self.Factors, int(len(self.Factors) / 6)):
                if first:
                    s += f'{self.Name} {self.Type} '
                    first = False
                else:
                    s += f'\n{self.Name} {" ":<{l}} '
                s += ' '.join([type2str(i) for i in a])
            return s
        else:
            return super().to_inp_line()


class Pollutant(BaseSectionObject):
    """
    Section: [**POLLUTANTS**]

    Purpose:
        Identifies the pollutants being analyzed.

    Format:
        ::

            Name Units Crain Cgw Cii Kd (Sflag CoPoll CoFract Cdwf Cinit)

    Format-PCSWMM:
        ``Name Units Crain Cgw Crdii Kdecay SnowOnly Co-Pollutant Co-Frac Cdwf Cinit``

    Remarks:
        ``FLOW`` is a reserved word and cannot be used to name a pollutant.

        Parameters Sflag through Cinit can be omitted if they assume their default values.
        If there is no co-pollutant but non-default values for Cdwf or Cinit, then enter an asterisk (``*``)
        for the co-pollutant name.

        When pollutant X has a co-pollutant Y, it means that fraction CoFract of pollutant Y's runoff
        concentration is added to pollutant X's runoff concentration when wash off from a subcatchment is computed.

        The dry weather flow concentration can be overridden for any specific node of the conveyance
        system by editing the node's Inflows property.

    Args:
        Name (str): name assigned to pollutant.
        unit (str): concentration units

                - ``MG/L`` for milligrams per liter
                - ``UG/L`` for micrograms per liter
                - ``#/L`` for direct count per liter

        Crain (float): concentration of pollutant in rainfall (concentration units).
        Cgw (float): concentration of pollutant in groundwater (concentration units).
        Crdii (float): concentration of pollutant in inflow/infiltration (concentration units). ``Cii``
        Kdecay (float): first-order decay coefficient (1/days).
        SnowOnly (bool): ``YES`` if pollutant buildup occurs only when there is snow cover, ``NO`` otherwise (default
        is ``NO``). ``Sflag``
        Co_Pollutant (str): name of co-pollutant (default is no co-pollutant designated by a ``*``). ``CoPoll``
        Co_Frac (float): fraction of co-pollutant concentration (default is 0). ``CoFract``
        Cdwf (float): pollutant concentration in dry weather flow (default is 0).
        Cinit (float): pollutant concentration throughout the conveyance system at the start of the simulation (
        default is 0).

    Attributes:
        Name (str): name assigned to pollutant.
        unit (str): concentration units

                - ``MG/L`` for milligrams per liter
                - ``UG/L`` for micrograms per liter
                - ``#/L`` for direct count per liter

        Crain (float): concentration of pollutant in rainfall (concentration units).
        Cgw (float): concentration of pollutant in groundwater (concentration units).
        Crdii (float): concentration of pollutant in inflow/infiltration (concentration units). ``Cii``
        Kdecay (float): first-order decay coefficient (1/days).
        SnowOnly (bool): ``YES`` if pollutant buildup occurs only when there is snow cover, ``NO`` otherwise (default
        is ``NO``). ``Sflag``
        Co_Pollutant (str): name of co-pollutant (default is no co-pollutant designated by a ``*``). ``CoPoll``
        Co_Frac (float): fraction of co-pollutant concentration (default is 0). ``CoFract``
        Cdwf (float): pollutant concentration in dry weather flow (default is 0).
        Cinit (float): pollutant concentration throughout the conveyance system at the start of the simulation (
        default is 0).
    """
    _identifier = IDENTIFIERS.Name
    _section_label = POLLUTANTS

    class UNITS:
        MG_PER_L = 'MG/L'
        UG_PER_L = 'UG/L'
        COUNT_PER_L = '#/L'

    def __init__(self, Name, unit, Crain, Cgw, Crdii, Kdecay,
                 SnowOnly=False, Co_Pollutant='*', Co_Frac=0, Cdwf=0, Cinit=0):
        self.Name = str(Name)
        self.unit = str(unit)
        self.Crain = float(Crain)
        self.Cgw = float(Cgw)
        self.Crdii = float(Crdii)
        self.Kdecay = float(Kdecay)
        self.SnowOnly = to_bool(SnowOnly)
        self.Co_Pollutant = str(Co_Pollutant)
        self.Co_Frac = float(Co_Frac)
        self.Cdwf = float(Cdwf)
        self.Cinit = float(Cinit)


class Transect(BaseSectionObject):
    """
    Section: [**TRANSECTS**]

    Purpose:
        Describes the cross-section geometry of natural channels or conduits with irregular shapes
        following the HEC-2 data format.

    Formats:

        ::

            NC Nleft Nright Nchanl
            X1 Name Nsta Xleft Xright 0 0 0 Lfactor Wfactor Eoffset
            GR Elev Station ... Elev Station

    Remarks:
        Transect geometry is described as shown below, assuming that one is looking in a downstream direction:

        The first line in this section must always be a NC line. After that, the NC line is only needed when a
        transect has
        different Manning’s n values than the previous one.

        The Manning’s n values on the NC line will supersede any roughness value entered for the conduit which uses the
        irregular cross-section.

        There should be one X1 line for each transect.
        Any number of GR lines may follow, and each GR line can have any number of Elevation-Station data pairs.
        (In HEC-2 the GR line is limited to 5 stations.)

        The station that defines the left overbank boundary on the X1 line must correspond to one of the station entries
        on the GR lines that follow. The same holds true for the right overbank boundary. If there is no match,
        a warning
        will be issued and the program will assume that no overbank area exists.

        The meander modifier is applied to all conduits that use this particular transect for their cross section.
        It assumes that the length supplied for these conduits is that of the longer main channel.
        SWMM will use the shorter overbank length in its calculations while increasing the main channel roughness to
        account
        for its longer length.

    Args:
        roughness_left (float): Manning’s n of right overbank portion of channel (use 0 if no change from previous NC
        line). ``Nleft``
        roughness_right (float): Manning’s n of right overbank portion of channel (use 0 if no change from previous
        NC line. ``Nright``
        roughness_channel (float): Manning’s n of main channel portion of channel (use 0 if no change from previous
        NC line. ``Nchanl``
        Name (str): name assigned to transect.
        bank_station_left (float): station position which ends the left overbank portion of the channel (ft or m).
        ``Xleft``
        bank_station_right (float): station position which begins the right overbank portion of the channel (ft or
        m). ``Xright``
        modifier_meander (float): meander modifier that represents the ratio of the length of a meandering main
        channel to the length of the overbank area that surrounds it (use 0 if not applicable). ``Lfactor``
        modifier_stations (float): factor by which distances between stations should be multiplied to increase (or
        decrease) the width of the channel (enter 0 if not applicable). ``Wfactor``
        modifier_elevations (float): amount added (or subtracted) from the elevation of each station (ft or m).
        ``Eoffset``
        station_elevations (list[list[float, float]]): of the tuple:

            Elev (float): elevation of the channel bottom at a cross-section station relative to some fixed reference
            (ft or m).
            Station (float): distance of a cross-section station from some fixed reference (ft or m).

    Attributes:
        roughness_left (float): Manning’s n of right overbank portion of channel (use 0 if no change from previous NC
        line). ``Nleft``
        roughness_right (float): Manning’s n of right overbank portion of channel (use 0 if no change from previous
        NC line. ``Nright``
        roughness_channel (float): Manning’s n of main channel portion of channel (use 0 if no change from previous
        NC line. ``Nchanl``
        Name (str): name assigned to transect.
        bank_station_left (float): station position which ends the left overbank portion of the channel (ft or m).
        ``Xleft``
        bank_station_right (float): station position which begins the right overbank portion of the channel (ft or
        m). ``Xright``
        modifier_meander (float): meander modifier that represents the ratio of the length of a meandering main
        channel to the length of the overbank area that surrounds it (use 0 if not applicable). ``Lfactor``
        modifier_stations (float): factor by which distances between stations should be multiplied to increase (or
        decrease) the width of the channel (enter 0 if not applicable). ``Wfactor``
        modifier_elevations (float): amount added (or subtracted) from the elevation of each station (ft or m).
        ``Eoffset``
        station_elevations (list[list[float, float]]): of the tuple:

            Elev (float): elevation of the channel bottom at a cross-section station relative to some fixed reference
            (ft or m).
            Station (float): distance of a cross-section station from some fixed reference (ft or m).
    """
    _identifier = IDENTIFIERS.Name
    _table_inp_export = False
    _section_label = TRANSECTS

    class KEYS:
        NC = 'NC'
        X1 = 'X1'
        GR = 'GR'

    def __init__(self, Name, station_elevations=None, bank_station_left=0, bank_station_right=0,
                 roughness_left=0, roughness_right=0, roughness_channel=0,
                 modifier_stations=0, modifier_elevations=0, modifier_meander=0):
        self.Name = str(Name)

        self.roughness_left = None
        self.roughness_right = None
        self.roughness_channel = None
        self.set_roughness(roughness_left, roughness_right, roughness_channel)

        self.bank_station_left = None
        self.bank_station_right = None
        self.set_bank_stations(bank_station_left, bank_station_right)

        self.modifier_stations = None
        self.modifier_elevations = None
        self.modifier_meander = None
        self.set_modifiers(modifier_meander, modifier_stations, modifier_elevations)

        self.station_elevations = list()

        if station_elevations is not None:
            for s in station_elevations:
                self.add_station_elevation(*s)

    def add_station_elevation(self, station, elevation):
        self.station_elevations.append([float(station), float(elevation)])

    def set_roughness(self, left=0., right=0., channel=0.):
        self.roughness_left = float(left)
        self.roughness_right = float(right)
        self.roughness_channel = float(channel)

    def set_bank_stations(self, left=0., right=0.):
        self.bank_station_left = float(left)
        self.bank_station_right = float(right)

    def set_modifiers(self, meander=0., stations=0., elevations=0.):
        self.modifier_stations = float(stations)
        self.modifier_elevations = float(elevations)
        self.modifier_meander = float(meander)

    def get_number_stations(self):
        """``Nsta`` number of stations across cross-section at which elevation data is supplied."""
        return len(self.station_elevations)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last_roughness = [0, 0, 0]
        last = None

        for line in multi_line_args:
            if line[0] == cls.KEYS.NC:
                last_roughness = line[1:]

            elif line[0] == cls.KEYS.X1:
                if last is not None:
                    yield last
                last = cls(Name=line[1])
                last.set_bank_stations(*line[3:5])
                last.set_modifiers(*line[8:])
                last.set_roughness(*last_roughness)

            elif line[0] == cls.KEYS.GR:
                it = iter(line[1:])
                for station in it:
                    elevation = next(it)
                    last.add_station_elevation(station, elevation)
        yield last

    def to_inp_line(self, break_every=1):
        """
        Args:
            break_every: break every x-th GR station, default: after every station
        """
        s = '{} {} {} {}\n'.format(self.KEYS.NC, self.roughness_left, self.roughness_right, self.roughness_channel)
        s += '{} {} {} {} {} 0 0 0 {} {} {}\n'.format(self.KEYS.X1, self.Name, self.get_number_stations(),
                                                      self.bank_station_left, self.bank_station_right,
                                                      self.modifier_meander, self.modifier_stations,
                                                      self.modifier_elevations, )
        if break_every == 1:
            for x, y in self.station_elevations:
                s += '{} {} {}\n'.format(self.KEYS.GR, x, y)
        else:
            s += self.KEYS.GR
            i = 0
            for x, y in self.station_elevations:
                s += ' {} {}'.format(x, y)
                i += 1
                if i == break_every:
                    i = 0
                    s += '\n' + self.KEYS.GR

        if s.endswith(self.KEYS.GR):
            s = s[:-3]
        s += '\n'
        return s


class Control(BaseSectionObject):
    """
    Section: [**CONTROLS**]

    Purpose:
        Determines how pumps and regulators will be adjusted
        based on simulation time or conditions at specific nodes and links.

    Formats:
        Each control rule is a series of statements of the form:
        ::

            RULE    ruleID
            IF      condition_1
            AND     condition_2
            OR      condition_3
            AND     condition_4
            Etc.
            THEN    action_1
            AND     action_2
            Etc.
            ELSE    action_3
            AND     action_4
            Etc.
            PRIORITY value

    Remarks:
        `RuleID`: an ID label assigned to the rule.
        `condition_n`: a condition clause.
        `action_n`: an action clause.
        `value`: a priority value (e.g., a number from 1 to 5).

        A **condition clause** of a Control Rule has the following format:
            `Object Name Attribute Relation Value`

        where:
            - `Object`: is a category of object,
            - `Name`: is the object’s assigned ID name,
            - `Attribute`: is the name of an attribute or property of the object,
            - `Relation`: is a relational operator (=, <>, <, <=, >, >=), and
            - `Value`: is an attribute value.

        for example:
        ::

            NODE N23 DEPTH > 10
            PUMP P45 STATUS = OFF
            SIMULATION TIME = 12:45:00
            NODE  N23  DEPTH  >  10
            NODE  N23  DEPTH  >  NODE 25 DEPTH
            PUMP  P45  STATUS =  OFF
            LINK  P45  TIMEOPEN >= 6:30
            SIMULATION CLOCKTIME = 22:45:00

        TIMEOPEN is the duration a link has been in an OPEN or ON state or have its
        SETTING be greater than zero; TIMECLOSED is the duration it has remained in a
        CLOSED or OFF state or have its SETTING be zero.

        +------------+------------+----------------------------------+
        | Object     | Attributes | Value                            |
        +============+============+==================================+
        | NODE       | DEPTH      | numerical value                  |
        +------------+------------+----------------------------------+
        |            | HEAD       | numerical value                  |
        +------------+------------+----------------------------------+
        |            | VOLUME     | numerical value                  |
        +------------+------------+----------------------------------+
        |            | INFLOW     | numerical value                  |
        +------------+------------+----------------------------------+
        | LINK       | FLOW       | numerical value                  |
        +------------+------------+----------------------------------+
        |            | DEPTH      | numerical value                  |
        +------------+------------+----------------------------------+
        |            | TIMEOPEN   | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        |            | TIMECLOSED | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        | CONDUIT    | STATUS     | OPEN or CLOSED                   |
        +------------+------------+----------------------------------+
        |            | TIMEOPEN   | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        |            | TIMECLOSED | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        | PUMP       | STATUS     | ON or OFF                        |
        +------------+------------+----------------------------------+
        |            | SETTING    | pump curve multiplier            |
        +------------+------------+----------------------------------+
        |            | FLOW       | numerical value                  |
        +------------+------------+----------------------------------+
        |            | TIMEOPEN   | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        |            | TIMECLOSED | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        | ORIFICE    | SETTING    | fraction open                    |
        +------------+------------+----------------------------------+
        |            | TIMEOPEN   | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        |            | TIMECLOSED | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        | WEIR       | SETTING    | fraction open                    |
        +------------+------------+----------------------------------+
        |            | TIMEOPEN   | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        |            | TIMECLOSED | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        | OUTLET     | SETTING    | rating curve multiplier          |
        +------------+------------+----------------------------------+
        |            | TIMEOPEN   | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        |            | TIMECLOSED | decimal hours or hr:min          |
        +------------+------------+----------------------------------+
        | SIMULATION | TIME       | elapsed time in decimal hours or |
        +------------+------------+----------------------------------+
        |            |            | hr:min:sec                       |
        +------------+------------+----------------------------------+
        | SIMULATION | DATE       | month/day/year                   |
        +------------+------------+----------------------------------+
        |            | MONTH      | month of year (January = 1)      |
        +------------+------------+----------------------------------+
        |            | DAY        | day of week (Sunday = 1)         |
        +------------+------------+----------------------------------+
        |            | CLOCKTIME  | time of day in hr:min:sec        |
        +------------+------------+----------------------------------+

        An **Action Clause** of a Control Rule has the following format:
            `Object Name Action = Value`
            `PUMP id STATUS = ON/OFF`
            `PUMP/ORIFICE/WEIR/OUTLET id SETTING = value`

        where:
            - `Object`: is a category of object,
            - `Name`: is the object’s assigned ID name,
            - `Action`: is the type of action
            - `Value`: is the action value/setting/status.

        where the meaning of SETTING depends on the object being controlled:
            - for Pumps it is a multiplier applied to the flow computed from the pump curve,
            - for Orifices it is the fractional amount that the orifice is fully open,
            - for Weirs it is the fractional amount of the original freeboard that exists (i.e., weir control is accomplished by moving the crest height up or down),
            - for Outlets it is a multiplier applied to the flow computed from the outlet's rating curve.

        Modulated controls are control rules that provide for a continuous degree of control
        applied to a pump or flow regulator as determined by the value of some controller
        variable, such as water depth at a node, or by time. The functional relation between
        the control setting and the controller variable is specified by using a control curve, a
        time series, or a PID controller. To model these types of controls, the value entry on
        the right-hand side of the action clause is replaced by the keyword CURVE,
        TIMESERIES, or PID and followed by the id name of the respective control curve or
        time series or by the gain, integral time (in minutes), and derivative time (in minutes)
        of a PID controller. For direct action control the gain is a positive number while for
        reverse action control it must be a negative number. By convention, the controller
        variable used in a control curve or PID control will always be the object and attribute
        named in the last condition clause of the rule. The value specified for this clause will
        be the setpoint used in a PID controller.


        for example:
        ::

            PUMP P67 STATUS = OFF
            ORIFICE O212 SETTING = 0.5
            WEIR W25 SETTING = CURVE C25
            ORIFICE ORI_23 SETTING = PID 0.1 0.1 0.0

        Only the RULE, IF and THEN portions of a rule are required; the other portions are optional.
        When mixing AND and OR clauses, the OR operator has higher precedence than AND, i.e., `IF A or B and C`
        is equivalent to `IF (A or B) and C.`
        If the interpretation was meant to be `IF A or (B and C)`
        then this can be expressed using two rules as in `IF A THEN ... IF B and C THEN ...`

        The PRIORITY value is used to determine which rule applies when two or more rules
        require that conflicting actions be taken on a link. A conflicting rule with a higher
        priority value has precedence over one with a lower value (e.g., PRIORITY 5
        outranks PRIORITY 1). A rule without a priority value always has a lower priority
        than one with a value. For two rules with the same priority value, the rule that
        appears first is given the higher priority.


    Examples:
        ::

            ; Simple time-based pump control
            RULE R1
            IF SIMULATION TIME > 8
            THEN PUMP 12 STATUS = ON
            ELSE PUMP 12 STATUS = OFF

            ; Multi-condition orifice gate control
            RULE R2A
            IF NODE 23 DEPTH > 12
            AND LINK 165 FLOW > 100
            THEN ORIFICE R55 SETTING = 0.5

            RULE R2B
            IF NODE 23 DEPTH > 12
            AND LINK 165 FLOW > 200
            THEN ORIFICE R55 SETTING = 1.0

            RULE R2C
            IF NODE 23 DEPTH <= 12
            OR LINK 165 FLOW <= 100
            THEN ORIFICE R55 SETTING = 0

            ; PID controller that attempts to keep Node 23’s depth at 12:
            RULE PID_1
            IF NODE 23 DEPTH <> 12
            THEN ORIFICE R55 SETTING = PID 0.5 0.1 0.0

            ; Pump station operation with a main (N1A) and lag (N1B) pump
            RULE R3A
            IF NODE N1 DEPTH > 5
            THEN PUMP N1A STATUS = ON

            RULE R3B
            IF PUMP N1A TIMEOPEN > 2:30
            THEN PUMP N1B STATUS = ON
            ELSE PUMP N1B STATUS = OFF

            RULE R3C
            IF NODE N1 DEPTH <= 0.5
            THEN PUMP N1A STATUS = OFF
            AND PUMP N1B STATUS = OFF

    """
    _identifier = IDENTIFIERS.Name
    _table_inp_export = False
    _section_label = CONTROLS

    class OBJECTS:
        NODE = 'NODE'
        LINK = 'LINK'
        CONDUIT = 'CONDUIT'
        PUMP = 'PUMP'
        ORIFICE = 'ORIFICE'
        WEIR = 'WEIR'
        OUTLET = 'OUTLET'
        SIMULATION = 'SIMULATION'

    class ATTRIBUTES:
        DEPTH = 'DEPTH'
        HEAD = 'HEAD'
        VOLUME = 'VOLUME'
        INFLOW = 'INFLOW'
        FLOW = 'FLOW'
        TIMEOPEN = 'TIMEOPEN'
        TIMECLOSED = 'TIMECLOSED'
        STATUS = 'STATUS'
        SETTING = 'SETTING'
        TIME = 'TIME'
        DATE = 'DATE'
        MONTH = 'MONTH'
        DAY = 'DAY'
        CLOCKTIME = 'CLOCKTIME'

    class LOGIC:
        RULE = 'RULE'
        IF = 'IF'
        THEN = 'THEN'
        PRIORITY = 'PRIORITY'
        AND = 'AND'
        OR = 'OR'

    class _Condition(BaseSectionObject):
        """
        A **condition clause** of a Control Rule has the following format:
            `Object Name Attribute Relation Value`
            `SIMULATION Attribute Relation Value`

        where:
            - `Object`: is a category of object,
            - `Name`: is the object’s assigned ID name,
            - `Attribute`: is the name of an attribute or property of the object,
            - `Relation`: is a relational operator (=, <>, <, <=, >, >=), and
            - `Value`: is an attribute value.

        for example:
        ::

            NODE N23 DEPTH > 10
            PUMP P45 STATUS = OFF
            SIMULATION TIME = 12:45:00
            NODE  N23  DEPTH  >  10
            NODE  N23  DEPTH  >  NODE 25 DEPTH
            PUMP  P45  STATUS =  OFF
            LINK  P45  TIMEOPEN >= 6:30
            SIMULATION CLOCKTIME = 22:45:00
        """
        def __init__(self, logic, kind, *args):
            self.logic = logic.upper()  # if and or
            self.kind = kind.upper()  # Control.OBJECTS
            self.label = NaN
            line = list(args)
            if kind.upper() == Control.OBJECTS.SIMULATION:
                pass
            else:
                self.label = line.pop(0)

            self.attribute = line.pop(0).upper()
            self.relation = line.pop(0)
            self.value = ' '.join(line)

    class _Action(BaseSectionObject):
        """
        An **Action Clause** of a Control Rule has the following format:
            `Object Name Action = Value`
            `PUMP id STATUS = ON/OFF`
            `PUMP/ORIFICE/WEIR/OUTLET id SETTING = value`

        where:
            - `Object`: is a category of object,
            - `Name`: is the object’s assigned ID name,
            - `Action`: is the type of action
            - `Value`: is the action value/setting/status.

        for example:
        ::

            PUMP P67 STATUS = OFF
            ORIFICE O212 SETTING = 0.5
            WEIR W25 SETTING = CURVE C25
            ORIFICE ORI_23 SETTING = PID 0.1 0.1 0.0
        """
        def __init__(self, logic, kind, label, action, relation, *value):
            self.logic = logic.upper()  # THAN, AND
            self.kind = kind.upper()  # Control.OBJECTS
            self.label = label
            self.action = action.upper()
            self.relation = relation  # immer "="  # always equal
            self.value = ' '.join(value)

    def __init__(self, Name, conditions, actions, priority=0):
        self.Name = str(Name)
        self.conditions = conditions
        self.actions = actions
        self.priority = int(priority)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        args = list()
        is_condition = False
        is_action = False
        for logic, *line in multi_line_args:
            if logic.upper() == cls.LOGIC.RULE:
                if args:
                    yield cls(*args)
                    args = list()
                args.append(line[0])
                is_action = False

            elif logic.upper() == cls.LOGIC.IF:
                args.append([cls._Condition(logic, *line)])
                is_condition = True

            elif logic.upper() == cls.LOGIC.THEN:
                args.append([cls._Action(logic, *line)])
                is_condition = False
                is_action = True

            elif logic.upper() == cls.LOGIC.PRIORITY:
                args.append(line[0])
                is_action = False

            elif is_condition:
                args[-1].append(cls._Condition(logic, *line))

            elif is_action:
                args[-1].append(cls._Action(logic, *line))

        # last
        yield cls(*args)

    def to_inp_line(self):
        s = f'{self.LOGIC.RULE} {self.Name}\n'
        s += '{}\n'.format('\n'.join([c.to_inp_line() for c in self.conditions]))
        s += '{}\n'.format('\n'.join([a.to_inp_line() for a in self.actions]))
        s += f'{self.LOGIC.PRIORITY} {self.priority}\n'
        return s


class Curve(BaseSectionObject):
    """
    Section: [**CURVES**]

    Purpose:
        Describes a relationship between two variables in tabular format.

    Format:
        ::

            Name Type X-value Y-value ...

    Format-PCSWMM:
            ``Name Type X-Value Y-Value``

    Remarks:
        Name
            name assigned to table
        Type
            ``STORAGE`` / ``SHAPE`` / ``DIVERSION`` / ``TIDAL`` / ``PUMP1`` / ``PUMP2`` / ``PUMP3`` / ``PUMP4`` /
            ``RATING`` / ``CONTROL``
        X-value
            an x (independent variable) value
        Y-value
            the y (dependent variable) value corresponding to x

        Multiple pairs of x-y values can appear on a line. If more than one line is needed,
        repeat the curve's name (but not the type) on subsequent lines. The x-values must
        be entered in increasing order.

        Choices for curve type have the following meanings (flows are expressed in the
        user’s choice of flow units set in the [``OPTIONS``] section):

            ``STORAGE``
                surface area in ft2 (m2) v. depth in ft (m) for a storage unit node
            ``SHAPE``
                width v. depth for a custom closed cross-section, both normalized with respect to full depth
            ``DIVERSION``
                diverted outflow v. total inflow for a flow divider node
            ``TIDAL``
                water surface elevation in ft (m) v. hour of the day for an outfall node
            ``PUMP1``
                pump outflow v. increment of inlet node volume in ft3 (m3)
            ``PUMP2``
                pump outflow v. increment of inlet node depth in ft (m)
            ``PUMP3``
                pump outflow v. head difference between outlet and inlet nodes in ft (m)
            ``PUMP4``
                pump outflow v. continuous depth in ft (m)
            ``RATING``
                outlet flow v. head in ft (m)
            ``CONTROL``
                control setting v. controller variable

    Examples:
        ::

            ;Storage curve (x = depth, y = surface area)
            AC1 STORAGE 0 1000 2 2000 4 3500 6 4200 8 5000
            ;Type1 pump curve (x = inlet wet well volume, y = flow )
            PC1 PUMP1
            PC1 100 5 300 10 500 20

    Args:
        Name (str): name assigned to table
        Type (str): one of ``STORAGE`` / ``SHAPE`` / ``DIVERSION`` / ``TIDAL`` / ``PUMP1`` / ``PUMP2`` / ``PUMP3`` /
        ``PUMP4`` / ``RATING`` / ``CONTROL``
        points (list[list[float, float]]): tuple of X-value (an independent variable) and  Y-value (an dependent
        variable)

    Attributes:
        Name (str): name assigned to table
        Type (str): one of ``STORAGE`` / ``SHAPE`` / ``DIVERSION`` / ``TIDAL`` / ``PUMP1`` / ``PUMP2`` / ``PUMP3`` /
        ``PUMP4`` / ``RATING`` / ``CONTROL``
        points (list[list[float, float]]): tuple of X-value (an independent variable) and  Y-value (an dependent
        variable)
    """
    _identifier = IDENTIFIERS.Name
    _table_inp_export = False
    _section_label = CURVES

    class TYPES:
        STORAGE = 'STORAGE'
        SHAPE = 'SHAPE'
        DIVERSION = 'DIVERSION'
        TIDAL = 'TIDAL'
        PUMP1 = 'PUMP1'
        PUMP2 = 'PUMP2'
        PUMP3 = 'PUMP3'
        PUMP4 = 'PUMP4'
        RATING = 'RATING'
        CONTROL = 'CONTROL'

    @classmethod
    def _get_names(cls, Type):
        TYPES = cls.TYPES
        if Type == TYPES.STORAGE:
            return ['depth', 'area']
        elif Type == TYPES.SHAPE:
            return ['depth', 'width']
        elif Type == TYPES.DIVERSION:
            return ['inflow', 'outflow']
        elif Type == TYPES.TIDAL:
            return ['hour', 'elevation']
        elif Type == TYPES.PUMP1:
            return ['volume', 'outflow']
        elif Type == TYPES.PUMP2:
            return ['depth', 'outflow']
        elif Type == TYPES.PUMP3:
            return ['head diff', 'outflow']
        elif Type == TYPES.PUMP4:
            return ['depth', 'outflow']
        elif Type == TYPES.RATING:
            return ['head', 'flow']
        elif Type == TYPES.CONTROL:
            return ['variable', 'setting']

    def __init__(self, Name, Type, points):
        self.Name = str(Name)
        self.Type = Type.upper()
        self.points = points

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last = None
        Type = None
        points = list()
        for name, *line in multi_line_args:
            remains = iter(line)

            if name != last:
                # new curve line
                if last is not None:
                    # first return previous curve
                    yield cls(last, Type, points)
                # reset variables
                points = list()
                last = name
                Type = next(remains)

            # points in current line
            for a in remains:
                b = next(remains)
                points.append(infer_type([a, b]))

        # last
        if last is not None:
            yield cls(last, Type, points)

    @property
    def frame(self):
        return DataFrame.from_records(self.points, columns=self._get_names(self.Type))

    def to_inp_line(self):
        points = iter(self.points)
        x, y = next(points)
        f = '{}  {} {:7.4f} {:7.4f}\n'.format(self.Name, self.Type, x, y)
        Type = ' ' * len(self.Type)
        for x, y in points:  # [(x,y), (x,y), ...]
            f += '{}  {} {:7.4f} {:7.4f}\n'.format(self.Name, Type, x, y)
        return f


class Timeseries(BaseSectionObject):
    """
    Section: [**TIMESERIES**]

    Purpose:
        Describes how a quantity varies over time.

    Formats:
        ::

            Name ( Date ) Hour Value ...
            Name Time Value ...
            Name FILE Fname

    Remarks:
        There are two options for supplying the data for a time series:
            i.: directly within this input file section as described by the first two formats
            ii.: through an external data file named with the third format.

        When direct data entry is used, multiple date-time-value or time-value entries can
        appear on a line. If more than one line is needed, the table's name must be repeated
        as the first entry on subsequent lines.

        When an external file is used, each line in the file must use the same formats listed
        above, except that only one date-time-value (or time-value) entry is allowed per line.
        Any line that begins with a semicolon is considered a comment line and is ignored.
        Blank lines are not allowed.

        Note that there are two methods for describing the occurrence time of time series data:

        - as calendar date/time of day (which requires that at least one date, at the start of the series, be entered)
        - as elapsed hours since the start of the simulation.

        For the first method, dates need only be entered at points in time when a new day occurs.

    Examples:
        ::

            ;Rainfall time series with dates specified
            TS1 6-15-2001 7:00 0.1 8:00 0.2 9:00 0.05 10:00 0
            TS1 6-21-2001 4:00 0.2 5:00 0 14:00 0.1 15:00 0

            ;Inflow hydrograph - time relative to start of simulation
            HY1 0 0 1.25 100 2:30 150 3.0 120 4.5 0
            HY1 32:10 0 34.0 57 35.33 85 48.67 24 50 0

    Args:
        Name (str): name assigned to time series.
    """
    _identifier = IDENTIFIERS.Name
    _table_inp_export = False
    _section_label = TIMESERIES

    class TYPES:
        FILE = 'FILE'

    def __init__(self, Name):
        self.Name = str(Name)

    @classmethod
    def _convert_lines(cls, multi_line_args):
        data = list()
        last = None
        last_date = None

        for name, *line in multi_line_args:
            # ---------------------------------
            if line[0].upper() == cls.TYPES.FILE:
                yield TimeseriesFile(name, ' '.join(line[1:]))
                last = name

            # ---------------------------------
            else:
                if name != last:
                    if last is not None:
                        yield TimeseriesData(last, data)
                    data = list()
                    last = name
                    last_date = None

                # -------------
                iterator = iter(line)
                for part in iterator:
                    if ('/' in part) or (('-' in part) and not part.startswith('-')):
                        # MM-DD-YYYY or MM/DD/YYYY or MMM-DD-YYYY MMM/DD/YYYY
                        last_date = part

                        # HH:MM or HH:MM:SS or H (as float)
                        time = next(iterator)
                    else:
                        # HH:MM or HH:MM:SS or H (as float)
                        time = part

                    if last_date is not None:
                        index = ' '.join([last_date, time])
                    else:
                        index = time

                    value = float(next(iterator))

                    data.append([index, value])

        # add last timeseries
        if data and (line[0].upper() != cls.TYPES.FILE):
            yield TimeseriesData(last, data)


class TimeseriesFile(Timeseries):
    """
    Section: [**TIMESERIES**]

    Purpose:
        Describes how a quantity varies over time.

    Formats:
        ::

            Name FILE Fname

    Args:
        Name (str): name assigned to time series.
        filename (str): name of a file in which the time series data are stored ``Fname``
    """

    def __init__(self, Name, filename, kind=None):
        Timeseries.__init__(self, Name)
        self.kind = self.TYPES.FILE
        self.filename = filename

    def to_inp_line(self):
        fn = self.filename.strip('"')
        return f'{self.Name} {self.kind} "{fn}"'


class TimeseriesData(Timeseries):
    """
    Section: [**TIMESERIES**]

    Purpose:
        Describes how a quantity varies over time.

    Formats:
        ::

            Name ( Date ) Hour Value ...
            Name Time Value ...

    Args:
        Name (str): name assigned to time series.
        data (list[tuple]): list of index/value tuple with:

            - Date: date in Month/Day/Year format (e.g., June 15, 2001 would be 6/15/2001).
            - Hour: 24-hour military time (e.g., 8:40 pm would be 20:40) relative to the last date specified
                   (or to midnight of the starting date of the simulation if no previous date was specified).
            - Time: hours since the start of the simulation, expressed as a decimal number or as hours:minutes.
            - Value: value corresponding to given date and time.
    """

    def __init__(self, Name, data):
        Timeseries.__init__(self, Name)
        self.data = data
        self._fix_index()

    def _fix_index(self):
        """
        convert string index to pandas.Timestamp (datetime like) or float

        index format supported by SWMM:
        - '%m/%d/%Y %H:%M'
        - '%b/%d/%Y %H:%M'
        - '%m-%d-%Y %H:%M'
        - '%b-%d-%Y %H:%M'
        - '%m/%d/%Y %H:%M:%S'
        - '%b/%d/%Y %H:%M:%S'
        - '%m-%d-%Y %H:%M:%S'
        - '%b-%d-%Y %H:%M:%S'
        - Hours as float relative to simulation start time
        """
        date_time, values = zip(*self.data)
        date_time_new = list()
        last_date = None
        if len(date_time) > 10000 * 2:  # 10000 it/s
            str_only = True
        else:
            str_only = False

        for dt in date_time:
            if isinstance(dt, Timestamp):
                date_time_new.append(dt)
            elif isinstance(dt, float):
                date_time_new.append(dt)
            else:
                parts = dt.split()
                if len(parts) == 1:
                    time = parts[0]
                else:
                    last_date, time = parts

                date_time_new.append(str_to_datetime(last_date, time, str_only=str_only))
        if str_only:
            import pandas as pd
            date_time_new = pd.to_datetime(date_time_new, format='%m/%d/%Y %H:%M:%S')
        self.data = list(zip(date_time_new, values))

    @property
    def frame(self):
        """
        convert object to pandas Series

        Returns:
            pandas.Series: Timeseries
        """
        datetime, values = zip(*self.data)
        return Series(index=datetime, data=values, name=self.Name)

    def to_inp_line(self):
        f = ''
        for date_time, value in self.data:
            f += f'{self.Name} {datetime_to_str(date_time)} {value}\n'
        return f

    @classmethod
    def from_pandas(cls, series, label=None):
        """
        convert pandas Series to TimeseriesData object

        Args:
            series (pandas.Series): timeseries with DateTimeIndex
                or index with correct formatted for SWMM:
                - '%m/%d/%Y %H:%M'
                - '%b/%d/%Y %H:%M'
                - '%m-%d-%Y %H:%M'
                - '%b-%d-%Y %H:%M'
                - '%m/%d/%Y %H:%M:%S'
                - '%b/%d/%Y %H:%M:%S'
                - '%m-%d-%Y %H:%M:%S'
                - '%b-%d-%Y %H:%M:%S'
                - Hours as float relative to simulation start time
            label (str): optional: label of the series. default: take series.name

        Returns:
            TimeseriesData: object for inp file
        """
        if label is None:
            label = series.name
        return cls(label, list(zip(series.index, series.values)))


class Tag(BaseSectionObject):
    """Section: [**TAGS**]"""
    _identifier = ('kind', IDENTIFIERS.Name)
    _section_label = TAGS

    class TYPES:
        Node = IDENTIFIERS.Node
        Subcatch = IDENTIFIERS.Subcatch
        Link = IDENTIFIERS.Link

    def __init__(self, kind, Name, tag, *tags):
        self.kind = kind.lower().capitalize()
        self.Name = Name
        self.tag = tag
        if tags:
            self.tag += '_' + '_'.join(tags)


class Label(BaseSectionObject):
    """
    Section: [**LABELS**]

    Purpose:
        Assigns X,Y coordinates to user-defined map labels.

    Format:
        ::

            Xcoord Ycoord Label (Anchor Font Size Bold Italic)

    Args:
        x (float):
            horizontal coordinate relative to origin in lower left of map.
        y (float):
            vertical coordinate relative to origin in lower left of map.
        label (str):
            text of label surrounded by double quotes.
        anchor (str):
            name of node or subcatchment that anchors the label on zoom-ins (use an empty pair of double quotes if
            there is no anchor).
        font (str):
            name of label’s font (surround by double quotes if the font name includes spaces).
        size (float):
            font size in points.
        bold (bool):
            YES for bold font, NO otherwise.
        italic (bool):
            YES for italic font, NO otherwise.
    """
    _identifier = ('x', 'y', 'label')
    _section_label = LABELS

    def __init__(self, x, y, label, anchor=NaN, font=NaN, size=NaN, bold=NaN, italic=NaN):
        self.x = float(x)
        self.y = float(y)
        self.label = label
        self.anchor = anchor
        self.font = font
        self.size = size
        self.bold = bold
        self.italic = italic

    @classmethod
    def from_inp_line(cls, *line):
        return cls(*split_line_with_quotes(line))


class Hydrograph(BaseSectionObject):
    """
    Section: [**HYDROGRAPHS**]

    Purpose:
        Specifies the shapes of the triangular unit hydrographs that determine the amount of
        rainfall-dependent infiltration/inflow (RDII) entering the drainage system.

    Formats:
        ::

            Name Raingage
            Name Month SHORT/MEDIUM/LONG R T K (Dmax Drec D0)

    Remarks:
        Name
            name assigned to a unit hydrograph group.
        Raingage
            name of the rain gage used by the unit hydrograph group.
        Month
            month of the year (e.g., JAN, FEB, etc. or ALL for all months).
        R
            response ratio for the unit hydrograph.
        T
            time to peak (hours) for the unit hydrograph.
        K
            recession limb ratio for the unit hydrograph.
        Dmax
            maximum initial abstraction depth available (in rain depth units).
        Drec
            initial abstraction recovery rate (in rain depth units per day)
        D0
            initial abstraction depth already filled at the start of the simulation (in rain depth units).

        For each group of unit hydrographs, use one line to specify its rain gage followed by
        as many lines as are needed to define each unit hydrograph used by the group
        throughout the year. Three separate unit hydrographs, that represent the short-term,
        medium-term, and long-term RDII responses, can be defined for each month (or all
        months taken together). Months not listed are assumed to have no RDII.

        The response ratio (R) is the fraction of a unit of rainfall depth that becomes RDII.
        The sum of the ratios for a set of three hydrographs does not have to equal 1.0.

        The recession limb ratio (K) is the ratio of the duration of the hydrograph’s recession
        limb to the time to peak (T) making the hydrograph time base equal to T*(1+K) hours.
        The area under each unit hydrograph is 1 inch (or mm).

        The optional initial abstraction parameters determine how much rainfall is lost at the
        start of a storm to interception and depression storage. If not supplied then the
        default is no initial abstraction.

    Examples:
        ::

            ; All three unit hydrographs in this group have the same shapes except those in July,
            ; which have only a short- and medium-term response and a different shape.
            UH101 RG1
            UH101 ALL SHORT 0.033 1.0 2.0
            UH101 ALL MEDIUM 0.300 3.0 2.0
            UH101 ALL LONG 0.033 10.0 2.0
            UH101 JUL SHORT 0.033 0.5 2.0
            UH101 JUL MEDIUM 0.011 2.0 2.0

    Args:
        Name (str): name assigned to time series.
    """
    _identifier = IDENTIFIERS.Name
    _table_inp_export = False
    _section_label = HYDROGRAPHS

    class TYPES:
        SHORT = 'SHORT'
        MEDIUM = 'MEDIUM'
        LONG = 'LONG'

    def __init__(self, Name, rain_gage, monthly_parameters=None):
        self.Name = str(Name)
        self.rain_gage = rain_gage

        if monthly_parameters is None:
            self.monthly_parameters = list()
        else:
            self.monthly_parameters = monthly_parameters

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last = None

        for name, *line in multi_line_args:
            # ---------------------------------
            if line[0].upper() not in cls.MONTHS._possible:
                if last is not None:
                    yield last
                last = cls(name, rain_gage=line[0])
            elif name == last.Name:
                last.monthly_parameters.append(cls.MONTHS.Parameters(name, *line))
        yield last

    class MONTHS:
        JAN = 'JAN'
        FEB = 'FEB'
        MAR = 'MAR'
        APR = 'APR'
        MAI = 'MAI'
        JUN = 'JUN'
        JUL = 'JUL'
        AUG = 'AUG'
        SEP = 'SEP'
        OCT = 'OCT'
        NOV = 'NOV'
        DEC = 'DEC'

        ALL = 'ALL'

        _possible = [JAN, FEB, MAR, APR, MAI, JUN, JUL, AUG, SEP, OCT, NOV, DEC, ALL]

        class Parameters(BaseSectionObject):
            _identifier = IDENTIFIERS.Name

            def __init__(self, Name, month, response, response_ratio, time_to_peak, recession_limb_ratio,
                         depth_max=NaN, depth_recovery=NaN, depth_init=NaN):
                """

                Args:
                    Name (str): name assigned to a unit hydrograph group.
                    month (str):
                    response (str):
                    response_ratio (float):
                    time_to_peak (float):
                    recession_limb_ratio (float):
                    depth_max (str):
                    depth_recovery (str):
                    depth_init (str):
                """
                self.Name = str(Name)
                self.month = month
                self.response = response
                self.response_ratio = float(response_ratio)
                self.time_to_peak = float(time_to_peak)
                self.recession_limb_ratio = float(recession_limb_ratio)
                self.depth_max = depth_max
                self.depth_recovery = depth_recovery
                self.depth_init = depth_init

    def to_inp_line(self):
        s = '{} {}\n'.format(self.Name, self.rain_gage)
        for hyd in self.monthly_parameters:
            s += hyd.to_inp_line() + '\n'
        return s


class LandUse(BaseSectionObject):
    """
    Section: [**LANDUSES**]

    Purpose:
        Identifies the various categories of land uses within the drainage area. Each
        subcatchment area can be assigned a different mix of land uses. Each land use can
        be subjected to a different street sweeping schedule.

    Formats:
        ::

            Name (SweepInterval Availability LastSweep)

    Args:
        Name:
            land use name.
        sweep_interval:
            days between street sweeping.
        availability:
            fraction of pollutant buildup available for removal by street sweeping.
        last_sweep:
            days since last sweeping at start of the simulation.
    """
    _identifier = IDENTIFIERS.Name
    _section_label = LANDUSES

    def __init__(self, Name, sweep_interval=NaN, availability=NaN, last_sweep=NaN):
        self.Name = str(Name)
        self.sweep_interval = float(sweep_interval)
        self.availability = float(availability)
        self.last_sweep = float(last_sweep)


class WashOff(BaseSectionObject):
    """
    Section: [**WASHOFF**]

    Purpose:
        Specifies the rate at which pollutants are washed off from different land uses during rain events.

    Formats:
        ::

            Landuse Pollutant FuncType C1 C2 SweepRmvl BmpRmvl

    Args:
        landuse:
            land use name.
        pollutant:
            pollutant name.
        func_type:
            washoff function type: ``EXP`` / ``RC`` / ``EMC``.
        C1 (float):
            washoff function coefficients(see Table D-3).
        C2 (float):
            washoff function coefficients(see Table D-3).
        sweeping_removal (float):
            street sweeping removal efficiency (percent) .
        BMP_removal (float):
            BMP removal efficiency (percent).

    Remarks:
        Table D-3 Pollutant wash off functions

        +------+--------------------------+--------------------------+------------+
        | Name | Function                 | Equation                 | Units      |
        +------+--------------------------+--------------------------+------------+
        | EXP  | Exponential              | C1 (runoff) C2 (buildup) | Mass/hour  |
        +------+--------------------------+--------------------------+------------+
        | RC   | Rating Curve             | C1 (runoff) C2           | Mass/sec   |
        +------+--------------------------+--------------------------+------------+
        | EMC  | Event Mean Concentration | C1                       | Mass/Liter |
        +------+--------------------------+--------------------------+------------+

        Each washoff function expresses its results in different units.

        For the Exponential function the runoff variable is expressed in catchment depth
        per unit of time (inches per hour or millimeters per hour), while for the Rating Curve
        function it is in whatever flow units were specified in the [``OPTIONS``] section of the
        input file (e.g., ``CFS``, ``CMS``, etc.).

        The buildup parameter in the Exponential function is the current total buildup over
        the subcatchment’s land use area in mass units. The units of C1 in the Exponential
        function are (in/hr)
        -C2 per hour (or (mm/hr) -C2 per hour). For the Rating Curve
        function, the units of ``C1`` depend on the flow units employed. For the EMC (event
        mean concentration) function, ``C1`` is always in concentration units.

    """
    _identifier = (IDENTIFIERS.Landuse, IDENTIFIERS.Pollutant)
    _section_label = WASHOFF

    class FUNCTIONS:
        EXP = 'EXP'
        RC = 'RC'
        EMC = 'EMC'

    def __init__(self, landuse, pollutant, func_type, C1, C2, sweeping_removal, BMP_removal):
        self.landuse = landuse
        self.pollutant = pollutant
        self.func_type = func_type
        self.C1 = float(C1)
        self.C2 = float(C2)
        self.sweeping_removal = float(sweeping_removal)
        self.BMP_removal = float(BMP_removal)


class BuildUp(BaseSectionObject):
    """
    Section: [**BUILDUP**]

    Purpose:
        Specifies the rate at which pollutants build up over different land uses between rain events.

    Formats:
        ::

            Landuse Pollutant FuncType C1 C2 C3 PerUnit

    Args:
        landuse:
            land use name.
        pollutant:
            pollutant name.
        func_type:
            buildup function type: ( ``POW`` / ``EXP`` / ``SAT`` / ``EXT`` ).
        C1 (float):
            buildup function parameters (see Table).
        C2 (float):
            buildup function parameters (see Table).
        C3 (float):
            buildup function parameters (see Table).
        per_unit (str):
            ``AREA`` if buildup is per unit area, ``CURBLENGTH`` if per length of curb.

    Remarks:
        Buildup is measured in pounds (kilograms) per unit of area (or curb length) for
        pollutants whose concentration units are either mg/L or ug/L. If the concentration
        units are counts/L, then the buildup is expressed as counts per unit of area (or curb
        length).

        Table: Pollutant buildup functions (t is antecedent dry days)

        +------+-------------+---------------------+
        | Name | Function    | Equation            |
        +------+-------------+---------------------+
        | POW  | Power       | Min (C1, C2*t^C3 )  |
        +------+-------------+---------------------+
        | EXP  | Exponential | C1*(1 – exp(-C2*t)) |
        +------+-------------+---------------------+
        | SAT  | Saturation  | C1*t / (C3 + t)     |
        +------+-------------+---------------------+
        | EXT  | External    | See below           |
        +------+-------------+---------------------+

        For the EXT buildup function, C1 is the maximum possible buildup (mass per area or
        curb length), C2 is a scaling factor, and C3 is the name of a Time Series that
        contains buildup rates (as mass per area or curb length per day) as a function of
        time.
    """
    _identifier = (IDENTIFIERS.Landuse, IDENTIFIERS.Pollutant)
    _section_label = BUILDUP

    class FUNCTIONS:
        EXP = 'EXP'
        RC = 'RC'
        EMC = 'EMC'

    class UNIT:
        AREA = 'AREA'
        CURBLENGTH = 'CURBLENGTH'
        CURB = 'CURB'

    def __init__(self, landuse, pollutant, func_type, C1, C2, C3, per_unit):
        self.landuse = landuse
        self.pollutant = pollutant
        self.func_type = func_type
        self.C1 = float(C1)
        self.C2 = float(C2)
        self.C3 = float(C3)
        self.per_unit = per_unit


class SnowPack(BaseSectionObject):
    """
    Section: [**SNOWPACKS**]

    Purpose:
        Specifies parameters that govern how snowfall accumulates and melts on the
        plowable, impervious and pervious surfaces of subcatchments.

    Formats:
        ::

            Name PLOWABLE Cmin Cmax Tbase FWF SD0 FW0 SNN0
            Name IMPERVIOUS Cmin Cmax Tbase FWF SD0 FW0 SD100
            Name PERVIOUS Cmin Cmax Tbase FWF SD0 FW0 SD100
            Name REMOVAL Dplow Fout Fimp Fperv Fimelt (Fsub Scatch)

    Args:
        Name (str):
            name assigned to snowpack parameter set.
        Cmin (float):
            minimum melt coefficient (in/hr-deg F or mm/hr-deg C).
        Cmax (float):
            maximum melt coefficient (in/hr-deg F or mm/hr-deg C).
        Tbase (float):
            snow melt base temperature (deg F or deg C).
        FWF (float):
            ratio of free water holding capacity to snow depth (fraction).
        SD0 (float):
            initial snow depth (in or mm water equivalent).
        FW0 (float):
            initial free water in pack (in or mm).
        SNN0 (float):
            fraction of impervious area that can be plowed.
        SD100 (float):
            snow depth above which there is 100% cover (in or mm water equivalent).
        Dplow (float):
            depth of snow on plowable areas at which snow removal begins (in or mm).
        Fout (float):
            fraction of snow on plowable area transferred out of watershed.
        Fimp (float):
            fraction of snow on plowable area transferred to impervious area by plowing.
        Fperv (float):
            fraction of snow on plowable area transferred to pervious area by plowing.
        Fimelt (float):
            fraction of snow on plowable area converted into immediate melt.
        Fsub (float):
            fraction of snow on plowable area transferred to pervious area in another subcatchment.
        Scatch (str):
            name of subcatchment receiving the Fsub fraction of transferred snow.

    Remarks:
        Use one set of PLOWABLE, IMPERVIOUS, and PERVIOUS lines for each snow pack
        parameter set created. Snow pack parameter sets are associated with specific
        subcatchments in the [SUBCATCHMENTS] section. Multiple subcatchments can share
        the same set of snow pack parameters.

        The PLOWABLE line contains parameters for the impervious area of a subcatchment
        that is subject to snow removal by plowing but not to areal depletion. This area is the
        fraction SNN0 of the total impervious area. The IMPERVIOUS line contains parameter
        values for the remaining impervious area and the PERVIOUS line does the same for
        the entire pervious area. Both of the latter two areas are subject to areal depletion.

        The REMOVAL line describes how snow removed from the plowable area is
        transferred onto other areas. The various transfer fractions should sum to no more
        than 1.0. If the line is omitted then no snow removal takes place.
    """
    _identifier = IDENTIFIERS.Name
    _section_label = SNOWPACKS
    _table_inp_export = False

    def __init__(self, Name, parts=None):
        self.Name = str(Name)
        self.parts = dict()

        if isinstance(parts, dict):
            self.parts = parts
        elif isinstance(parts, list):
            for p in parts:
                self.add_pack(p)
        elif parts is None:
            pass
        else:
            raise NotImplementedError(f'SnowPack packs type "{type(parts)}" not implemented!')

    def add_pack(self, p):
        if isinstance(p, self.PARTS._possible_types):
            self.parts[p._LABEL] = p

    @classmethod
    def _convert_lines(cls, multi_line_args):
        last = None

        for name, kind, *line in multi_line_args:
            # ---------------------------------
            if last is None:
                last = cls(name)

            elif name != last.Name:
                yield last
                last = cls(name)

            kind = kind.upper()
            last.parts[kind] = cls.PARTS._dict[kind](*line)
        yield last

    class PARTS:
        class _Base(BaseSectionObject):
            _table_inp_export = False
            _identifier = IDENTIFIERS.Name
            _section_label = SNOWPACKS

            def __init__(self, Cmin, Cmax, Tbase, FWF, SD0, FW0):
                self.Cmin = float(Cmin)
                self.Cmax = float(Cmax)
                self.Tbase = float(Tbase)
                self.FWF = float(FWF)
                self.SD0 = float(SD0)
                self.FW0 = float(FW0)

        class Plowable(_Base):
            _LABEL = 'PLOWABLE'

            def __init__(self, Cmin, Cmax, Tbase, FWF, SD0, FW0, SNN0):
                SnowPack.PARTS._Base.__init__(self, Cmin, Cmax, Tbase, FWF, SD0, FW0)
                self.SNN0 = float(SNN0)

        class Pervious(_Base):
            _LABEL = 'PERVIOUS'

            def __init__(self, Cmin, Cmax, Tbase, FWF, SD0, FW0, SD100):
                SnowPack.PARTS._Base.__init__(self, Cmin, Cmax, Tbase, FWF, SD0, FW0)
                self.SD100 = float(SD100)

        class Impervious(Pervious):
            _LABEL = 'IMPERVIOUS'

        class Removal(BaseSectionObject):
            _LABEL = 'REMOVAL'

            def __init__(self, Dplow, Fout, Fimp, Fperv, Fimelt, Fsub=NaN, Scatch=NaN):
                self.Dplow = float(Dplow)
                self.Fout = float(Fout)
                self.Fimp = float(Fimp)
                self.Fperv = float(Fperv)
                self.Fimelt = float(Fimelt)
                self.Fsub = float(Fsub)
                self.Scatch = Scatch

        PLOWABLE = Plowable._LABEL
        IMPERVIOUS = Pervious._LABEL
        PERVIOUS = Impervious._LABEL
        REMOVAL = Removal._LABEL

        _possible_types = (Plowable, Pervious, Impervious, Removal)
        _possible = (PLOWABLE, PERVIOUS, IMPERVIOUS, REMOVAL)

        _dict = {x._LABEL: x for x in _possible_types}

    @property
    def plowable(self):
        return self.parts[self.PARTS.PLOWABLE]

    @property
    def impervious(self):
        return self.parts[self.PARTS.IMPERVIOUS]

    @property
    def pervious(self):
        return self.parts[self.PARTS.PERVIOUS]

    @property
    def removal(self):
        return self.parts[self.PARTS.REMOVAL]

    def to_inp_line(self):
        s = ''
        for pack in self.PARTS._possible:
            if self.parts[pack] is not None:
                s += f'{self.Name} {pack:<8} {self.parts[pack].to_inp_line()}\n'
        return s


class Aquifer(BaseSectionObject):
    """
    Section: [**AQUIFERS**]

    Purpose:
        Supplies parameters for each unconfined groundwater aquifer in the study area.
        Aquifers consist of two zones – a lower saturated zone and an upper unsaturated
        zone with a moving boundary between the two.

    Formats:
        ::

            Name Por WP FC Ks Kslp Tslp ETu ETs Seep Ebot Egw Umc (Epat)

    Args:
        Name (str):
            name assigned to aquifer.
        Por (float):
            soil porosity (volumetric fraction).
        WP (float):
            soil wilting point (volumetric fraction).
        FC (float):
            soil field capacity (volumetric fraction).
        Ks (float):
            saturated hydraulic conductivity (in/hr or mm/hr).
        Kslp (float):
            slope of the logarithm of hydraulic conductivity versus moisture deficit (i.e., porosity minus moisture
            content) curve (in/hr or mm/hr).
        Tslp (float):
            slope of soil tension versus moisture content curve (inches or mm).
        ETu (float):
            fraction of total evaporation available for evapotranspiration in the upper unsaturated zone.
        ETs (float):
            maximum depth into the lower saturated zone over which evapotranspiration can occur (ft or m).
        Seep (float):
            seepage rate from saturated zone to deep groundwater when water table is at ground surface (in/hr or mm/hr).
        Ebot (float):
            elevation of the bottom of the aquifer (ft or m).
        Egw (float):
            groundwater table elevation at start of simulation (ft or m).
        Umc (float):
            unsaturated zone moisture content at start of simulation (volumetric fraction).
        Epat (float):
            name of optional monthly time pattern used to adjust the upper zone evaporation fraction for different
            months of the year.

    Remarks:
        Local values for ``Ebot``, ``Egw``, and ``Umc`` can be assigned to specific subcatchments in
        the [``GROUNDWATER``] section described below.
    """
    _identifier = IDENTIFIERS.Name
    _section_label = AQUIFERS

    def __init__(self, Name, Por, WP, FC, Ks, Kslp, Tslp, ETu, ETs, Seep, Ebot, Egw, Umc, Epat=NaN):
        self.Name = str(Name)
        self.Por = float(Por)
        self.WP = float(WP)
        self.FC = float(FC)
        self.Ks = float(Ks)
        self.Kslp = float(Kslp)
        self.Tslp = float(Tslp)
        self.ETu = float(ETu)
        self.ETs = float(ETs)
        self.Seep = float(Seep)
        self.Ebot = float(Ebot)
        self.Egw = float(Egw)
        self.Umc = float(Umc)
        self.Epat = float(Epat)



import numpy as np


def section_from_frame(df, section_class):
    """
    create a inp-file section from an pandas DataFrame

    will only work for specific sections ie. JUNCTIONS where every row is the same and no special __init__ is needed!

    Args:
        df (pandas.DataFrame): data
        section_class (BaseSectionObject):

    Returns:
        InpSection: converted section
    """
    # TODO: macro_snippets
    values = np.vstack((df.index.values, df.values.T)).T
    return section_class.create_section(values)
    # return cls.from_lines([line.split() for line in dataframe_to_inp_string(df).split('\n')], section_class)


def group_edit(inp):
    # ToDo
    # for objects of type (Subcatchments, Infiltration, Junctions, Storage Units, or Conduits)
    # with tag equal to
    # edit the property
    # by replacing it with
    pass


def short_status(inp):
    for section in inp:
        print(f'{section}: {len(inp[section])}')


"""PCSWMM Simplify network tool 
Criteria

The criteria is a list of attributes and a given tolerance for each. Attribute criteria can be toggled on or off 
by checking the box next to the attribute. The attribute criteria are as follows:

    Cross-section shapes must match exactly
    Diameter values match within a specified percent tolerance
    Roughness values match within a specified percent tolerance
    Slope values match within a specified tolerance
    Transects must match exactly
    Shape curves must match exactly

Join preference

Select a preference for joining two conduits. The preference is used when two conduits are available for 
connecting on the upstream and downstream nodes. Choose the priority for the connection to be any one of the 
following:

    shorter conduit
    longer conduit
    closest slope
    upstream conduit
    downstream conduit
"""


# def dissolve_node(inp, node):
#     """
#     delete node and combine conduits
#
#     Args:
#         inp (InpData): inp data
#         node (str | Junction | Storage | Outfall): node to delete
#
#     Returns:
#         InpData: inp data
#     """
#     if isinstance(node, str):
#         node = find_node(inp, node)
#     # create new section with only
#     c1 = inp[sec.CONDUITS].slice_section([node.Name], 'ToNode')
#     if c1:
#         c2 = inp[sec.CONDUITS].slice_section([node.Name], 'FromNode')
#         inp = combine_conduits(inp, c1, c2)
#     else:
#         inp = delete_node(node.Name)
#     return inp



class OBJECTS:
    """
    All different types of objects in .out-file

    Attributes:
        SUBCATCHMENT (str): "subcatchment"
        NODE (str): "node"
        LINK (str): "link"
        POLLUTANT (str): "pollutant"
        SYSTEM (str): "system"
        LIST_ (list[str]): list of all objects
    """
    SUBCATCHMENT = "subcatchment"
    NODE = "node"
    LINK = "link"
    POLLUTANT = "pollutant"
    SYSTEM = "system"

    LIST_ = [SUBCATCHMENT, NODE, LINK, POLLUTANT, SYSTEM]


class SUBCATCHMENT_VARIABLES:
    """
    All variable names for sub-catchments

    Attributes:
        RAINFALL (str): "rainfall"
        SNOW_DEPTH (str): "snow_depth"
        EVAPORATION (str): "evaporation"
        INFILTRATION (str): "infiltration"
        RUNOFF (str): "runoff"
        GW_OUTFLOW (str): "groundwater_outflow"
        GW_ELEVATION (str): "groundwater_elevation"
        SOIL_MOISTURE (str): "soil_moisture"
        LIST_ (list[str]): list of all variables
    """
    RAINFALL = "rainfall"  # rainfall intensity | Rainfall rate [mm/hr] | Precipitation
    SNOW_DEPTH = "snow_depth"  # SNOWDEPTH | snow depth [mm]
    EVAPORATION = "evaporation"  # EVAP | evap loss [mm/day] | Evaporation
    INFILTRATION = "infiltration"  # INFIL | infil loss [mm/h] | Infiltration
    RUNOFF = "runoff"  # RUNOFF | runoff flow rate
    GW_OUTFLOW = "groundwater_outflow"  # GW_FLOW | groundwater flow rate to node | Groundwater flow into the drainage network
    GW_ELEVATION = "groundwater_elevation"  # GW_ELEV | elevation of saturated gw table [m]
    SOIL_MOISTURE = "soil_moisture"  # SOIL_MOIST | Soil moisture in the unsaturated groundwater zone [fraction]
    # Concentration (mass/volume) Washoff concentration of each pollutant. | TSS

    LIST_ = [RAINFALL, SNOW_DEPTH, EVAPORATION, INFILTRATION, RUNOFF, GW_OUTFLOW, GW_ELEVATION, SOIL_MOISTURE]


class NODE_VARIABLES:
    """
    All variable names for nodes

    Attributes:
        DEPTH (str): 'depth'
        HEAD (str): 'head'
        VOLUME (str): 'volume'
        LATERAL_INFLOW (str): 'lateral_inflow'
        TOTAL_INFLOW (str): 'total_inflow'
        FLOODING (str): 'flooding'
        LIST_ (list[str]): list of all variables
    """
    DEPTH = 'depth'  # DEPTH | water depth above invert | Water depth above the node invert elevation [m]
    HEAD = 'head'  # HEAD | hydraulic head | Absolute elevation per vertical datum [m]
    VOLUME = 'volume'  # VOLUME | volume stored & ponded | Stored water volume including ponded water [m³]
    LATERAL_INFLOW = 'lateral_inflow'  # LATFLOW | lateral inflow rate | Runoff + all other external (DW+GW+I&I+User) inflows
    TOTAL_INFLOW = 'total_inflow'  # INFLOW | total inflow rate | Lateral inflow + upstream inflows
    FLOODING = 'flooding'  # OVERFLOW | overflow rate | Flooding | Surface flooding; excess overflow when the node is at full depth
    # CONCENTRATION [mass/volume] Concentration of each pollutant after any treatment. | TSS

    LIST_ = [DEPTH, HEAD, VOLUME, LATERAL_INFLOW, TOTAL_INFLOW, FLOODING]


class LINK_VARIABLES:
    """
    All variable names for links

    Attributes:
        FLOW (str): 'flow'
        DEPTH (str): 'depth'
        VELOCITY (str): 'velocity'
        VOLUME (str): 'volume'
        CAPACITY (str): 'capacity'
        LIST_ (list[str]): list of all variables
    """
    FLOW = 'flow'  # FLOW | flow rate
    DEPTH = 'depth'  # DEPTH | flow depth | Average water depth [m]
    VELOCITY = 'velocity'  # VELOCITY | flow velocity
    VOLUME = 'volume'  # VOLUME | link volume |Volume of water in the conduit; this is based on the midpoint depth and midpoint cross sectional area [m³]
    CAPACITY = 'capacity'  # CAPACITY | ratio of area to full area |Fraction of full area filled by flow for conduits; control setting for pumps and regulators.
    # Concentration (mass/volume) Concentration of each pollutant. | TSS

    LIST_ = [FLOW, DEPTH, VELOCITY, VOLUME, CAPACITY]


class SYSTEM_VARIABLES:
    """
    All variable names system-wide values

    Attributes:
        AIR_TEMPERATURE (str): 'air_temperature'
        RAINFALL (str): 'rainfall'
        SNOW_DEPTH (str): 'snow_depth'
        INFILTRATION (str): 'infiltration'
        RUNOFF (str): 'runoff'
        DW_INFLOW (str): 'dry_weather_inflow'
        GW_INFLOW (str): 'groundwater_inflow'
        RDII_INFLOW (str): 'RDII_inflow'
        DIRECT_INFLOW (str): 'direct_inflow'
        LATERAL_INFLOW (str): 'lateral_inflow'
        FLOODING (str): 'flooding'
        OUTFLOW (str): 'outflow'
        VOLUME (str): 'volume'
        EVAPORATION (str): 'evaporation'
        PET (str): 'PET'
        LIST_ (list[str]): list of all variables
    """
    # somewhere are Losses by Exfiltration in STORAGES ??

    AIR_TEMPERATURE = 'air_temperature'  # TEMPERATURE | air temperature [°C]
    RAINFALL = 'rainfall'  # RAINFALL | rainfall intensity | Total rainfall [mm/hr] | Precipitation
    SNOW_DEPTH = 'snow_depth'  # SNOWDEPTH | snow depth | Total snow depth [mm]
    INFILTRATION = 'infiltration'  # INFIL | infil | Average system losses [mm/hr] | Infiltration # sum in infiltration of all sub-catchments (weighted by the sc-area)
    RUNOFF = 'runoff'  # RUNOFF | runoff flow | Total runoff flow
    DW_INFLOW = 'dry_weather_inflow'  # DWFLOW | dry weather inflow | Total dry weather inflow | DW Inflow
    GW_INFLOW = 'groundwater_inflow'  # GWFLOW | ground water inflow | Total groundwater inflow | GW Inflow
    RDII_INFLOW = 'RDII_inflow'  # IIFLOW | RDII inflow | Total rainfall derived infiltration and inflow (RDII). | I&I Inflow
    DIRECT_INFLOW = 'direct_inflow'  # EXFLOW | external inflow | Total direct inflow | Direct Inflow # inflow directly defined in section [INFLOWS] and assigned to node
    LATERAL_INFLOW = 'lateral_inflow'  # INFLOW | total lateral inflow | total external inflow | Total Inflow # INFLOW = RUNOFF + DWFLOW + GWFLOW + IIFLOW + EXFLOW
    FLOODING = 'flooding'  # FLOODING | flooding outflow | Total external flooding
    OUTFLOW = 'outflow'  # OUTFLOW | outfall outflow | Total outflow from outfalls
    VOLUME = 'volume'  # STORAGE | storage volume | Total nodal storage volume in the system [m³]
    EVAPORATION = 'evaporation'  # EVAP | evaporation | Actual evaporation [mm/day]
    PET = 'PET'  # PET | potential ET | Potential evaporation [mm/day]

    LIST_ = [AIR_TEMPERATURE, RAINFALL, SNOW_DEPTH, INFILTRATION, RUNOFF, DW_INFLOW,
             GW_INFLOW, RDII_INFLOW, DIRECT_INFLOW, LATERAL_INFLOW, FLOODING,
             OUTFLOW, VOLUME, EVAPORATION, PET]


# from swmmtoolbox | swmm source | swmm gui | swmm manual
# https://support.chiwater.com/77882/swmm5-output-file
class VARIABLES:
    """
    All different variables of the objects in the .out-file

    Attributes:
        SUBCATCHMENT (swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES): All variable-names for subcatchments
        NODE (swmm_api.output_file.definitions.NODE_VARIABLES): All variable-names for nodes
        LINK (swmm_api.output_file.definitions.LINK_VARIABLES): All variable-names for links
        SYSTEM (swmm_api.output_file.definitions.SYSTEM_VARIABLES): All variable-names for system values
    """
    SUBCATCHMENT = SUBCATCHMENT_VARIABLES
    NODE = NODE_VARIABLES
    LINK = LINK_VARIABLES
    SYSTEM = SYSTEM_VARIABLES



"""
Python package:
https://github.com/fhirschmann/rdp

ist aber nicht schnell (Faktor 10 langsamer)

https://en.wikipedia.org/wiki/Ramer–Douglas–Peucker_algorithm

Code von hier:
https://stackoverflow.com/questions/2573997/reduce-number-of-points-in-line#10934629
"""


def _vec2d_dist(p1, p2):
    return (p1[0] - p2[0])**2 + (p1[1] - p2[1])**2


def _vec2d_sub(p1, p2):
    return p1[0]-p2[0], p1[1]-p2[1]


def _vec2d_mult(p1, p2):
    return p1[0]*p2[0] + p1[1]*p2[1]


def ramer_douglas(line, dist):
    """Does Ramer-Douglas-Peucker simplification of a curve with `dist`
    threshold.

    `line` is a list-of-tuples, where each tuple is a 2D coordinate

    Usage is like so:

    >>> myline = [(0.0, 0.0), (1.0, 2.0), (2.0, 1.0)]
    >>> simplified = ramer_douglas(myline, dist = 1.0)
    """

    if len(line) < 3:
        return line

    begin = line[0]
    end = line[-1] if line[0] != line[-1] else line[-2]

    distSq = list()
    for curr in line[1:-1]:
        tmp = (
            _vec2d_dist(begin, curr) - _vec2d_mult(_vec2d_sub(end, begin), _vec2d_sub(curr, begin)) ** 2 / _vec2d_dist(begin, end))
        distSq.append(tmp)

    maxdist = max(distSq)
    if maxdist < dist ** 2:
        return [begin, end]

    pos = distSq.index(maxdist)
    return (ramer_douglas(line[:pos + 2], dist) +
            ramer_douglas(line[pos + 1:], dist)[1:])



from ..section_labels import JUNCTIONS, CONDUITS, STORAGE, OUTFALLS, ORIFICES
from ..sections import Storage, Outfall, Orifice


def junction_to_storage(inp, label, *args, **kwargs):
    """
    convert :class:`~swmm_api.input_file.inp_sections.node.Junction` to
    :class:`~swmm_api.input_file.inp_sections.node.Storage`

    and add it to the STORAGE section

    Args:
        inp (SwmmInput): inp-file data
        label (str): label of the junction
        *args: argument of the :class:`~swmm_api.input_file.inp_sections.node.Storage`-class
        **kwargs: keyword arguments of the :class:`~swmm_api.input_file.inp_sections.node.Storage`-class
    """
    j = inp[JUNCTIONS].pop(label)  # type: Junction
    if STORAGE not in inp:
        inp[STORAGE] = Storage.create_section()
    inp[STORAGE].add_obj(Storage(Name=label, Elevation=j.Elevation, MaxDepth=j.MaxDepth,
                                 InitDepth=j.InitDepth, Apond=j.Aponded, *args, **kwargs))


def junction_to_outfall(inp, label, *args, **kwargs):
    """
    convert :class:`~swmm_api.input_file.inp_sections.node.Junction` to
    :class:`~swmm_api.input_file.inp_sections.node.Outfall`

    and add it to the OUTFALLS section

    Args:
        inp (SwmmInput): inp-file data
        label (str): label of the junction
        *args: argument of the :class:`~swmm_api.input_file.inp_sections.node.Outfall`-class
        **kwargs: keyword arguments of the :class:`~swmm_api.input_file.inp_sections.node.Outfall`-class
    """
    j = inp[JUNCTIONS].pop(label)  # type: Junction
    if OUTFALLS not in inp:
        inp[OUTFALLS] = Outfall.create_section()
    inp[OUTFALLS].add_obj(Outfall(Name=label, Elevation=j.Elevation, *args, **kwargs))


def conduit_to_orifice(inp, label, Type, Offset, Qcoeff, FlapGate=False, Orate=0):
    """
    convert :class:`~swmm_api.input_file.inp_sections.link.Conduit` to
    :class:`~swmm_api.input_file.inp_sections.link.Orifice`

    and add it to the ORIFICES section

    Args:
        inp (SwmmInput): inp-file data
        label (str): label of the conduit
        Type (str): orientation of orifice: either SIDE or BOTTOM.
        Offset (float): amount that a Side Orifice’s bottom or the position of a Bottom Orifice is offset above
            the invert of inlet node (ft or m, expressed as either a depth or as an elevation,
            depending on the LINK_OFFSETS option setting).
        Qcoeff (float): discharge coefficient (unitless).
        FlapGate (bool): YES if flap gate present to prevent reverse flow, NO if not (default is NO).
        Orate (int): time in decimal hours to open a fully closed orifice (or close a fully open one).
                        Use 0 if the orifice can open/close instantaneously.
    """
    c = inp[CONDUITS].pop(label)  # type: Conduit
    if ORIFICES not in inp:
        inp[ORIFICES] = Orifice.create_section()
    inp[ORIFICES].add_obj(Orifice(Name=label, FromNode=c.FromNode, ToNode=c.ToNode,
                                  Type=Type, Offset=Offset, Qcoeff=Qcoeff, FlapGate=FlapGate, Orate=Orate))



from networkx import DiGraph
from numpy import interp, mean, ceil

from .collection import nodes_dict, links_dict, subcatchments_per_node_dict
from .graph import next_links_labels, previous_links, previous_links_labels, links_connected
from .macros import calc_slope, find_link, delete_sections
from ..inp import SwmmInput
from ..section_labels import *
from ..section_lists import NODE_SECTIONS, LINK_SECTIONS, SUBCATCHMENT_SECTIONS, POLLUTANT_SECTIONS
from ..sections import Tag, DryWeatherFlow, Junction, Coordinate, Conduit, Loss, Vertices, EvaporationSection


def delete_node(inp: SwmmInput, node_label, graph: DiGraph = None, alt_node=None):
    """
    delete node in inp data

    Args:
        inp (SwmmInput): inp data
        node_label (str): label of node to delete
        graph (DiGraph): networkx graph of model
        alt_node (str): node label | optional: move flows to this node

    .. Important::
        works inplace
    """
    for section in NODE_SECTIONS + [COORDINATES]:
        if (section in inp) and (node_label in inp[section]):
            inp[section].pop(node_label)

    if (TAGS in inp) and ((Tag.TYPES.Node, node_label) in inp.TAGS):
        del inp[TAGS][(Tag.TYPES.Node, node_label)]

    # AND delete connected links
    if graph is not None:
        if node_label in graph:
            links = next_links_labels(graph, node_label) + previous_links_labels(graph, node_label)  # type: List[str]
            graph.remove_node(node_label)
        else:
            links = list()
    else:
        links = list()
        for section in LINK_SECTIONS:
            if section in inp:
                links += list(inp[section].filter_keys([node_label], by='FromNode')) + \
                         list(inp[section].filter_keys([node_label], by='ToNode'))  # type: List[Conduit]
        links = [l.Name for l in links]  # type: List[str]

    for link in links:
        delete_link(inp, link)

    if alt_node is not None:
        move_flows(inp, node_label, alt_node)


def move_flows(inp: SwmmInput, from_node, to_node, only_constituent=None):
    """
    move flow (INFLOWS or DWF) from one node to another

    Args:
        inp (SwmmInput): inp data
        from_node (str): first node label
        to_node (str): second node label
        only_constituent (list): only consider this constituent (default: FLOW)

    Notes:
        works inplace
    """
    for section in (INFLOWS, DWF):
        if section not in inp:
            continue

        if only_constituent is None:
            only_constituent = [DryWeatherFlow.TYPES.FLOW]

        for constituent in only_constituent:
            index_old = (from_node, constituent)

            if index_old in inp[section]:
                index_new = (to_node, constituent)

                obj = inp[section].pop(index_old)
                obj.Node = to_node

                if index_new not in inp[section]:
                    inp[section].add_obj(obj)

                elif section == DWF:
                    # DryWeatherFlow can be easily added when Patterns are equal
                    inp[section][index_new].Base += obj.Base

                    # if not all([old[p] == new[p] for p in ['pattern1', 'pattern2', 'pattern3', 'pattern4']]):
                    #     print(f'WARNING: move_flows  from "{from_node}" to "{to_node}". DWF patterns don\'t
                    #     match!')

                elif section == INFLOWS:
                    # Inflows can't be added due to the multiplication factor / timeseries
                    # if (TimeSeries, Type, Mfactor, Sfactor,Pattern) are equal then sum(Baseline)
                    print(f'WARNING: move_flows  from "{from_node}" to "{to_node}". Already Exists!')

            # else:
            #     print(f'Nothing to move from "{from_node}" [{section}]')


def delete_link(inp: SwmmInput, link):
    for s in LINK_SECTIONS + [XSECTIONS, LOSSES, VERTICES]:
        if (s in inp) and (link in inp[s]):
            inp[s].pop(link)

    if (TAGS in inp) and ((Tag.TYPES.Link, link) in inp.TAGS):
        del inp[TAGS][(Tag.TYPES.Link, link)]


def delete_subcatchment(inp: SwmmInput, subcatchment):
    for s in SUBCATCHMENT_SECTIONS + [LOADINGS, COVERAGES]:
        # , GWF, GROUNDWATER
        if (s in inp) and (subcatchment in inp[s]):
            inp[s].pop(subcatchment)

    if (TAGS in inp) and ((Tag.TYPES.Subcatch, subcatchment) in inp.TAGS):
        del inp[TAGS][(Tag.TYPES.Subcatch, subcatchment)]


def split_conduit(inp, conduit, intervals=None, length=None, from_inlet=True):
    # mode = [cut_point (GUI), intervals (n), length (l)]
    nodes = nodes_dict(inp)
    if isinstance(conduit, str):
        conduit = inp[CONDUITS][conduit]  # type: Conduit

    dx = 0
    n_new_nodes = 0
    if intervals:
        dx = conduit.Length / intervals
        n_new_nodes = intervals - 1
    elif length:
        dx = length
        n_new_nodes = ceil(conduit.Length / length - 1)

    from_node = nodes[conduit.FromNode]
    to_node = nodes[conduit.ToNode]

    from_node_coord = inp[COORDINATES][from_node.Name]
    to_node_coord = inp[COORDINATES][to_node.Name]

    loss = None
    if (LOSSES in inp) and (conduit.Name in inp[LOSSES]):
        loss = inp[LOSSES][conduit.Name]  # type: Loss

    new_nodes = list()
    new_links = list()

    x = dx
    last_node = from_node
    for new_node_i in range(n_new_nodes + 1):
        if x >= conduit.Length:
            node = to_node
        else:
            node = Junction(Name=f'{from_node.Name}_{to_node.Name}_{chr(new_node_i + 97)}',
                            Elevation=interp(x, [0, conduit.Length], [from_node.Elevation, to_node.Elevation]),
                            MaxDepth=interp(x, [0, conduit.Length], [from_node.MaxDepth, to_node.MaxDepth]),
                            InitDepth=interp(x, [0, conduit.Length], [from_node.InitDepth, to_node.InitDepth]),
                            SurDepth=interp(x, [0, conduit.Length], [from_node.SurDepth, to_node.SurDepth]),
                            Aponded=float(mean([from_node.Aponded, to_node.Aponded])),
                            )
            new_nodes.append(node)
            inp[JUNCTIONS].add_obj(node)

            # TODO: COORDINATES based on vertices
            inp[COORDINATES].add_obj(Coordinate(node.Name,
                                                    x=interp(x, [0, conduit.Length],
                                                                [from_node_coord.x, to_node_coord.x]),
                                                    y=interp(x, [0, conduit.Length],
                                                                [from_node_coord.y, to_node_coord.y])))

        link = Conduit(Name=f'{conduit.Name}_{chr(new_node_i + 97)}',
                       FromNode=last_node.Name,
                       ToNode=node.Name,
                       Length=dx,
                       Roughness=conduit.Roughness,
                       InOffset=0 if new_node_i != 0 else conduit.InOffset,
                       OutOffset=0 if new_node_i != (n_new_nodes - 1) else conduit.OutOffset,
                       InitFlow=conduit.InitFlow,
                       MaxFlow=conduit.MaxFlow)
        new_links.append(link)
        inp[CONDUITS].add_obj(link)

        xs = inp[XSECTIONS][conduit.Name].copy()
        xs.Link = link.Name
        inp[XSECTIONS].add_obj(xs)

        if loss:
            inlet = loss.Inlet if loss.Inlet and (new_node_i == 0) else 0
            outlet = loss.Outlet if loss.Outlet and (new_node_i == n_new_nodes - 1) else 0
            average = loss.Average / (n_new_nodes + 1)
            flap_gate = loss.FlapGate

            if any([inlet, outlet, average, flap_gate]):
                inp[LOSSES].add_obj(Loss(link.Name, inlet, outlet, average, flap_gate))

        # TODO: VERTICES

        if node is to_node:
            break
        last_node = node
        x += dx

    # if conduit.Name in inp[VERTICES]:
    #     pass
    # else:
    #     # interpolate coordinates
    #     pass

    delete_link(inp, conduit.Name)


def combine_vertices(inp: SwmmInput, label1, label2):
    if COORDINATES not in inp:
        # if there are not coordinates this function is nonsense
        return

    vertices_class = Vertices

    if VERTICES not in inp:
        # we will at least ad the coordinates of the common node
        inp[VERTICES] = Vertices.create_section()
    else:
        vertices_class = inp[VERTICES]._section_object

    new_vertices = list()

    if label1 in inp[VERTICES]:
        new_vertices += list(inp[VERTICES][label1].vertices)

    common_node = links_dict(inp)[label1].ToNode
    if common_node in inp[COORDINATES]:
        new_vertices += [inp[COORDINATES][common_node].point]

    if label2 in inp[VERTICES]:
        new_vertices += list(inp[VERTICES][label2].vertices)

    if label1 in inp[VERTICES]:
        inp[VERTICES][label1].vertices = new_vertices
    else:
        inp[VERTICES].add_obj(vertices_class(label1, vertices=new_vertices))


def combine_conduits(inp, c1, c2, graph: DiGraph = None):
    """
    combine the two conduits to one keep attributes of the first (c1)

    Args:
        inp (SwmmInput): inp data
        c1 (str | Conduit): conduit 1 to combine
        c2 (str | Conduit): conduit 2 to combine
        graph (networkx.DiGraph): optional, runs faster with graph (inp representation)

    Returns:
        Conduit: new combined conduit

    .. Important::
        works inplace
    """
    if isinstance(c1, str):
        c1 = inp[CONDUITS][c1]
    if isinstance(c2, str):
        c2 = inp[CONDUITS][c2]
    # -------------------------
    if graph:
        graph.remove_edge(c1.FromNode, c1.ToNode)
    # -------------------------
    if c1.FromNode == c2.ToNode:
        c_first = c2.copy()  # type: Conduit
        c_second = c1.copy()  # type: Conduit
    elif c1.ToNode == c2.FromNode:
        c_first = c1.copy()  # type: Conduit
        c_second = c2.copy()  # type: Conduit
    else:
        raise EnvironmentError('Links not connected')

    # -------------------------
    # vertices + Coord of middle node
    combine_vertices(inp, c_first.Name, c_second.Name)

    # -------------------------
    c_new = c1  # type: Conduit
    # -------------------------
    common_node = c_first.ToNode
    c_new.FromNode = c_first.FromNode
    c_new.ToNode = c_second.ToNode
    # -------------------------
    if graph:
        graph.add_edge(c_new.FromNode, c_new.ToNode, label=c_new.Name)

    if isinstance(c_new, Conduit):
        c_new.Length = round(c1.Length + c2.Length, 1)

        # offsets
        c_new.InOffset = c_first.InOffset
        c_new.OutOffset = c_second.OutOffset

    # Loss
    if (LOSSES in inp) and (c_new.Name in inp[LOSSES]):
        print(f'combine_conduits {c1.Name} and {c2.Name}. BUT WHAT TO DO WITH LOSSES?')
        # add losses
        pass

    delete_node(inp, common_node, graph=graph, alt_node=c_new.FromNode)
    return c_new


def combine_conduits_keep_slope(inp, c1, c2, graph: DiGraph = None):
    nodes = nodes_dict(inp)
    new_out_offset = (- calc_slope(inp, c1) * c2.Length
                      + c1.OutOffset
                      + nodes[c1.ToNode].Elevation
                      - nodes[c2.ToNode].Elevation)
    c1 = combine_conduits(inp, c1, c2, graph=graph)
    c1.OutOffset = round(new_out_offset, 2)
    return c1


def dissolve_conduit(inp, c: Conduit, graph: DiGraph = None):
    """
    combine the two conduits to one

    Args:
        inp (SwmmInput): inp data
        c1 (str | Conduit): conduit 1 to combine
        c2 (str | Conduit): conduit 2 to combine
        keep_first (bool): keep first (of conduit 1) cross-section; else use second (of conduit 2)

    Returns:
        SwmmInput: inp data
    """
    common_node = c.FromNode
    for c_old in list(previous_links(inp, common_node, g=graph)):
        if graph:
            graph.remove_edge(c_old.FromNode, c_old.ToNode)

        c_new = c_old  # type: Conduit

        # vertices + Coord of middle node
        combine_vertices(inp, c_new.Name, c.Name)

        c_new.ToNode = c.ToNode
        # -------------------------
        if graph:
            graph.add_edge(c_new.FromNode, c_new.ToNode, label=c_new.Name)

        # Loss
        if LOSSES in inp and c_new.Name in inp[LOSSES]:
            print(f'dissolve_conduit {c.Name} in {c_new.Name}. BUT WHAT TO DO WITH LOSSES?')

        if isinstance(c_new, Conduit):
            c_new.Length = round(c.Length + c_new.Length, 1)
            # offsets
            c_new.OutOffset = c.OutOffset

    delete_node(inp, common_node, graph=graph, alt_node=c.ToNode)


def rename_node(inp: SwmmInput, old_label: str, new_label: str, g=None):
    """
    change node label

    Args:
        inp (SwmmInput): inp data
        old_label (str): previous node label
        new_label (str): new node label
        g (DiGraph): optional - graph of the network

    .. Important::
        works inplace
        CONTROLS Not Implemented!
    """
    # ToDo: Not Implemented: CONTROLS

    # Nodes and basic node components
    for section in NODE_SECTIONS + [COORDINATES, RDII]:
        if (section in inp) and (old_label in inp[section]):
            inp[section][new_label] = inp[section].pop(old_label)
            if hasattr(inp[section][new_label], 'Name'):
                inp[section][new_label].Name = new_label
            else:
                inp[section][new_label].Node = new_label

    # tags
    if (TAGS in inp) and ((Tag.TYPES.Node, old_label) in inp.TAGS):
        tag = inp[TAGS].pop((Tag.TYPES.Node, old_label))
        tag.Name = new_label
        inp.TAGS.add_obj(tag)

    # subcatchment outlets
    if SUBCATCHMENTS in inp:
        for obj in subcatchments_per_node_dict(inp)[old_label]:
            obj.Outlet = new_label
        # -------
        # for obj in inp.SUBCATCHMENTS.filter_keys([old_label], 'Outlet'):  # type: SubCatchment
        #     obj.Outlet = new_label
        # -------

    # link: from-node and to-node
    previous_links, next_links = links_connected(inp, old_label, g=g)
    for link in previous_links:
        link.ToNode = new_label

    for link in next_links:
        link.ToNode = new_label

    # -------
    # for section in [CONDUITS, PUMPS, ORIFICES, WEIRS, OUTLETS]:
    #     if section in inp:
    #         for obj in inp[section].filter_keys([old_label], 'FromNode'):  # type: _Link
    #             obj.FromNode = new_label
    #
    #         for obj in inp[section].filter_keys([old_label], 'ToNode'):  # type: _Link
    #             obj.ToNode = new_label
    # -------

    # (dwf-)inflows
    constituents = [DryWeatherFlow.TYPES.FLOW]
    if POLLUTANTS in inp:
        constituents += list(inp.POLLUTANTS.keys())

    for section in [INFLOWS, DWF, TREATMENT]:
        if section in inp:
            for constituent in constituents:
                old_id = (old_label, constituent)
                if old_id in inp[section]:
                    inp[section][old_id].Node = new_label
                    inp[section][(new_label, constituent)] = inp[section].pop(old_id)

            # -------
            # for obj in inp[section].filter_keys([old_label], 'Node'):  # type: Inflow
            #     obj.Node = new_label
            #     inp[section][(new_label, obj[obj._identifier[1]])] = inp[section].pop((old_label,
            #     obj[obj._identifier[1]]))
            # -------


def rename_link(inp: SwmmInput, old_label: str, new_label: str):
    """
    change link label

    Notes:
        works inplace
        CONTROLS Not Implemented!

    Args:
        inp (SwmmInput): inp data
        old_label (str): previous link label
        new_label (str): new link label
    """
    # ToDo: Not Implemented: CONTROLS
    for section in LINK_SECTIONS + [XSECTIONS, LOSSES, VERTICES]:
        if (section in inp) and (old_label in inp[section]):
            inp[section][new_label] = inp[section].pop(old_label)
            if hasattr(inp[section][new_label], 'Name'):
                inp[section][new_label].Name = new_label
            else:
                inp[section][new_label].Link = new_label

    if (TAGS in inp) and ((Tag.TYPES.Link, old_label) in inp.TAGS):
        inp[TAGS][(Tag.TYPES.Link, new_label)] = inp[TAGS].pop((Tag.TYPES.Link, old_label))


def rename_subcatchment(inp: SwmmInput, old_label: str, new_label: str):
    for section in SUBCATCHMENT_SECTIONS + [LOADINGS, COVERAGES, GWF, GROUNDWATER]:
        if (section in inp) and (old_label in inp[section]):
            inp[section][new_label] = inp[section].pop(old_label)
            if hasattr(inp[section][new_label], 'Name'):
                inp[section][new_label].Name = new_label
            else:
                inp[section][new_label].Subcatch = new_label

    if (TAGS in inp) and ((Tag.TYPES.Subcatch, old_label) in inp.TAGS):
        inp[TAGS][(Tag.TYPES.Subcatch, new_label)] = inp[TAGS].pop((Tag.TYPES.Subcatch, old_label))


def rename_timeseries(inp, old_label, new_label):
    """
    change timeseries label

    Args:
        inp (SwmmInput): inp data
        old_label (str): previous timeseries label
        new_label (str): new timeseries label

    .. Important::
        works inplace
    """
    if old_label in inp[TIMESERIES]:
        obj = inp[TIMESERIES].pop(old_label)
        obj.Name = new_label
        inp[TIMESERIES].add_obj(obj)

    key = EvaporationSection.KEYS.TIMESERIES  # TemperatureSection.KEYS.TIMESERIES, ...

    if RAINGAGES in inp:
        f = inp[RAINGAGES].frame
        filtered_table = f[(f['Source'] == key) & (f['Timeseries'] == old_label)]
        if not filtered_table.empty:
            for i in filtered_table.index:
                inp[RAINGAGES][i].Timeseries = new_label

    if EVAPORATION in inp:
        if key in inp[EVAPORATION]:
            if inp[EVAPORATION][key] == old_label:
                inp[EVAPORATION][key] = new_label

    if TEMPERATURE in inp:
        if key in inp[TEMPERATURE]:
            if inp[TEMPERATURE][key] == old_label:
                inp[TEMPERATURE][key] = new_label

    if OUTFALLS in inp:
        f = inp[OUTFALLS].frame
        filtered_table = f[(f['Type'] == key) & (f['Data'] == old_label)]
        if not filtered_table.empty:
            for i in filtered_table.index:
                inp[OUTFALLS][i].Data = new_label

    if INFLOWS in inp:
        f = inp[INFLOWS].frame
        filtered_table = f[f['TimeSeries'] == old_label]
        if not filtered_table.empty:
            for i in filtered_table.index:
                inp[INFLOWS][i].TimeSeries = new_label


def flip_link_direction(inp, link_label):
    link = find_link(inp, link_label)
    if link:
        link.FromNode, link.ToNode = link.ToNode, link.FromNode


def remove_quality_model(inp):
    """
    remove all sections only for modelling quality

    Args:
        inp (SwmmInput): inp-data

    .. Important::
        works inplace
    """
    delete_sections(inp, POLLUTANT_SECTIONS)

    for sec in [INFLOWS, DWF]:
        for k in list(inp[sec].keys()):
            if inp[sec][k].Constituent != 'FLOW':
                del inp[sec][k]



import abc
import struct
from io import SEEK_SET
from os import remove

_RECORDSIZE = 4


class BinaryReader(abc.ABC):
    def __init__(self, filename):
        self.fp = None
        self.fp = open(filename, "rb")

    @abc.abstractmethod
    def __repr__(self):
        return f'SwmmOutExtract(file="{self.filename}")'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Close the .out-file.

        Close the binary file to prevent memory errors.
        """
        if self.fp is not None:
            self.fp.close()

    def delete(self):
        """
        Close and delete the .out-file.
        """
        self.close()
        remove(self.filename)

    def __del__(self):
        self.close()

    @property
    def filename(self):
        """
        path and filename of the .out-file

        Returns:
            str: path and filename of the .out-file
        """
        return self.fp.name

    def _set_position(self, offset, whence=SEEK_SET):
        if self.fp.tell() != offset:
            self.fp.seek(offset, whence)

    def _next(self, n=1, dtype='i', flat=True):
        """
        read the next number or string in the binary .out-file

        Args:
            n (int): number of objects to read
            dtype (str): type of number to read ['i', 'f', 'd', 's', ...]
            flat (float): if just one object should be returned

        Returns:
            (int | float | str): read object
        """
        size = {'d': 2}.get(dtype, 1)
        if dtype == 's':
            s = self._next_base(f'{n}s', n)[0]
            return str(s, encoding="ascii", errors="replace")
        elif flat and (n == 1):
            return self._next_base(dtype, size * _RECORDSIZE)[0]
        else:
            return self._next_base(f'{n}{dtype}', n * size * _RECORDSIZE)

    def _next_base(self, fmt, size):
        """
        read from the binary .out-file

        Args:
            fmt (str): format of the read object(s)
            size (int): bytes to read

        Returns:
            (int | float | str): read object
        """
        return struct.unpack(fmt, self.fp.read(size))

    def _next_float(self):
        return self._next_base('f', _RECORDSIZE)[0]

    def _next_floats(self,  n):
        return self._next_base(f'{n}f', n * _RECORDSIZE)

    def _next_double(self):
        return self._next_base('d', _RECORDSIZE * 2)[0]

    def _next_doubles(self,  n):
        return self._next_base(f'{n}d', 2 * n * _RECORDSIZE)



from matplotlib.pyplot import subplots
from numpy import append, arange, array, concatenate


def curve_figure(curve):
    """
    create a plot for a given curve

    Args:
        curve (swmm_api.input_file.sections.Curve):

    Returns:
        matplotlib.pyplot.Figure: curve plot
    """
    fig, ax = subplots()

    # -------------------------
    xlim = 1
    ylim = 1
    base = 0.2
    xlim_left = -xlim

    # -------------------------
    y_label, x_label = curve._get_names(curve.Type)
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    ax.set_title('{}: {}'.format(curve.Name, curve.Type))

    # -------------------------
    y, x = list(zip(*curve.points))

    y = array(y)
    x = array(x) / 2

    y = concatenate(([0], y, [1]))
    x = concatenate(([0], x, [0]))

    y = append(y, y[::-1])
    x = append(x, x[::-1] * -1)

    # -------------------------
    ax.plot(x, y, marker='.', ls='-', zorder=1000000, clip_on=False)

    # -------------------------
    # ax.legend().remove()
    ax.set_aspect('equal', 'box')
    ax.set_xticks(arange(xlim_left, xlim, base), minor=False)
    if base / 2 != 0:
        ax.set_xticks(arange(xlim_left, xlim, base / 2), minor=True)

    ax.set_yticks(arange(0, ylim, base), minor=False)
    if base / 2 != 0:
        ax.set_yticks(arange(0, ylim, base / 2), minor=True)

    ax.tick_params(which='both', length=0, width=0, labelbottom=False, labeltop=False, labelleft=False,
                   labelright=False, bottom=False, top=False, left=False, right=False)

    ax.set_xlim(xlim_left, xlim)
    ax.set_ylim(0, ylim)
    ax.grid(True)
    # ax.grid(True, which='minor', linestyle=':', linewidth=0.5)
    ax.set_axisbelow(True)

    # ------------------
    fig.tight_layout()
    return fig



from _warnings import warn
# from os import path, remove
import os

from pandas import to_datetime

from swmm_api.input_file.macros import (find_node,
                                        find_link, calc_slope, conduit_iter_over_inp, combined_subcatchment_frame, )
from swmm_api.input_file.macros.convert import junction_to_storage, junction_to_outfall
from swmm_api.input_file.macros.reduce_unneeded import reduce_curves, reduce_raingages
from swmm_api.input_file.macros.edit import delete_node, combine_conduits
from swmm_api import read_inp_file
from swmm_api.input_file.section_types import SECTION_TYPES
from swmm_api.input_file import SwmmInput, section_labels as sec
from swmm_api.input_file._type_converter import offset2delta
from swmm_api.output_file import parquet
from swmm_api.output_file.out import read_out_file
from swmm_api.run import swmm5_run


class InpMacros(SwmmInput):
    def __init__(self):
        SwmmInput.__init__(self, {})
        self.filename = None
        self.basename = None
        self.dirname = None
        self._out = None

    def set_name(self, name):
        self.filename = name
        self.basename = '.'.join(os.path.basename(name).split('.')[:-1])
        self.dirname = os.path.dirname(name)

    @property
    def report_filename(self):
        return os.path.join(self.dirname, self.basename + '.rpt')

    @property
    def out_filename(self):
        return os.path.join(self.dirname, self.basename + '.out')

    @property
    def parquet_filename(self):
        return os.path.join(self.dirname, self.basename + '.parquet')

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.to_string()

    def read_file(self, **kwargs):
        data = read_inp_file(self.filename, **kwargs)
        SwmmInput.__init__(self, data)

    @classmethod
    def from_file(cls, filename, **kwargs):
        inp = cls()
        inp.set_name(filename)
        inp.read_file(**kwargs)
        return inp

    def write(self, fast=True):
        self.write_file(self.filename, fast=fast)

    # @classmethod
    # def from_pickle(cls, fn):
    #     new = cls()
    #     pkl_file = open(fn, 'rb')
    #     new._data = pickle.load(pkl_file)
    #     pkl_file.close()
    #     return new
    #
    # def to_pickle(self, fn):
    #     output = open(fn, 'wb')
    #     pickle.dump(self._data, output)
    #     output.close()

    # ------------------------------------------------------------------------------------------------------------------
    def execute_swmm(self, rpt_dir=None, out_dir=None, init_print=False):
        swmm5_run(self.filename, rpt_dir=rpt_dir, out_dir=out_dir, init_print=init_print)

    def delete_report_file(self):
        os.remove(self.report_filename)

    def delete_inp_file(self):
        os.remove(self.filename)

    def run(self, rpt_dir=None, out_dir=None, init_print=False):
        self.execute_swmm(rpt_dir=rpt_dir, out_dir=out_dir, init_print=init_print)
        self.convert_out()

    def __getitem__(self, section):
        self.check_section(section)
        return SwmmInput.__getitem__(self, section)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def output_data(self):
        if self._out is None:
            self._out = read_out_file(self.out_filename)
        return self._out

    def get_out_frame(self):
        return self.output_data.to_frame()

    def convert_out(self):
        self.output_data.to_parquet()
        self.delete_out_file()

    def delete_out_file(self):
        # TODO check if file can be deleted
        self._out.close()
        try:
            os.remove(self.out_filename)
        except PermissionError as e:
            warn(str(e))

    def get_result_frame(self):
        if not os.path.isfile(self.parquet_filename):
            data = self.output_data.to_frame()
            self.convert_out()
            return data
        else:
            return parquet.read(self.parquet_filename)

    ####################################################################################################################
    def reset_section(self, section):
        if section in self:
            del self[section]
        self[section] = SECTION_TYPES[section]

    def check_section(self, section):
        if section not in self:
            self[section] = SECTION_TYPES[section]

    def set_start(self, start):
        if isinstance(start, str):
            start = to_datetime(start)
        self[sec.OPTIONS]['START_DATE'] = start.date()
        self[sec.OPTIONS]['START_TIME'] = start.time()

    def set_start_report(self, start):
        if isinstance(start, str):
            start = to_datetime(start)
        self[sec.OPTIONS]['REPORT_START_DATE'] = start.date()
        self[sec.OPTIONS]['REPORT_START_TIME'] = start.time()

    def set_end(self, end):
        if isinstance(end, str):
            end = to_datetime(end)
        self[sec.OPTIONS]['END_DATE'] = end.date()
        self[sec.OPTIONS]['END_TIME'] = end.time()

    def set_threads(self, num):
        self[sec.OPTIONS]['THREADS'] = num

    def ignore_rainfall(self, on=True):
        self[sec.OPTIONS]['IGNORE_RAINFALL'] = on

    def ignore_snowmelt(self, on=True):
        self[sec.OPTIONS]['IGNORE_SNOWMELT'] = on

    def ignore_groundwater(self, on=True):
        self[sec.OPTIONS]['IGNORE_GROUNDWATER'] = on

    def ignore_quality(self, on=True):
        self[sec.OPTIONS]['IGNORE_QUALITY'] = on

    def set_intervals(self, freq):
        new_step = offset2delta(freq)
        self[sec.OPTIONS]['REPORT_STEP'] = new_step
        self[sec.OPTIONS]['WET_STEP'] = new_step
        self[sec.OPTIONS]['DRY_STEP'] = new_step

    def activate_report(self, input=False, continuity=True, flowstats=True, controls=False):
        self[sec.REPORT]['INPUT'] = input
        self[sec.REPORT]['CONTINUITY'] = continuity
        self[sec.REPORT]['FLOWSTATS'] = flowstats
        self[sec.REPORT]['CONTROLS'] = controls

    def add_obj_to_report(self, obj_kind, new_obj):
        if isinstance(new_obj, str):
            new_obj = [new_obj]
        elif isinstance(new_obj, list):
            pass
        else:
            raise NotImplementedError('Type: {} not implemented!'.format(type(new_obj)))

        old_obj = self[sec.REPORT][obj_kind]
        if isinstance(old_obj, str):
            old_obj = [old_obj]
        elif isinstance(old_obj, (int, float)):
            old_obj = [str(old_obj)]
        elif isinstance(old_obj, list):
            pass
        elif old_obj is None:
            old_obj = []
        else:
            raise NotImplementedError('Type: {} not implemented!'.format(type(old_obj)))

        self[sec.REPORT][obj_kind] = old_obj + new_obj

    def add_nodes_to_report(self, new_nodes):
        self.add_obj_to_report('NODES', new_nodes)

    def add_links_to_report(self, new_links):
        self.add_obj_to_report('LINKS', new_links)

    # def add_timeseries_file(self, fn):  # TODO
    #     if 'Files' not in self[sec.TIMESERIES]:
    #         self[sec.TIMESERIES]['Files'] = DataFrame(columns=['Type', 'Fname'])
    #
    #     self[sec.TIMESERIES]['Files'] = self[sec.TIMESERIES]['Files'].append(
    #         Series({'Fname': '"' + fn + '.dat"'}, name=path.basename(fn)))
    #     self[sec.TIMESERIES]['Files']['Type'] = 'FILE'

    def reduce_curves(self):
        reduce_curves(self)

    def reduce_raingages(self):
        reduce_raingages(self)

    def combined_subcatchment_infos(self):
        return combined_subcatchment_frame(self)

    def find_node(self, label):
        return find_node(self, label)

    def find_link(self, label):
        return find_link(self, label)

    def calc_slope(self, link_label):
        return calc_slope(self, link_label)

    def delete_node(self, node_label):
        delete_node(self, node_label)

    def combine_conduits(self, c1, c2):
        combine_conduits(self, c1, c2)

    def conduit_iter_over_inp(self, start, end=None):
        conduit_iter_over_inp(self, start, end=end)

    def junction_to_outfall(self, label, *args, **kwargs):
        junction_to_outfall(self, label, *args, **kwargs)

    def junction_to_storage(self, label, *args, **kwargs):
        junction_to_storage(self, label, *args, **kwargs)



import shape_generator
from .macros import find_link
from ..sections import CrossSection, Pump
from ..inp import SwmmInput


VIRTUAL_LENGTH = 100


def get_cross_section_maker(inp: SwmmInput, link_label: str):
    c = find_link(inp, link_label)
    if c is None:
        return  # not found
    if isinstance(c, Pump):
        return
    xs = inp.XSECTIONS[link_label]

    if xs.Shape == CrossSection.SHAPES.CUSTOM:
        curve = inp.CURVES[xs.Curve]
        return shape_generator.CrossSection.from_curve(curve, height=VIRTUAL_LENGTH)
    elif xs.Shape == CrossSection.SHAPES.IRREGULAR:
        return  # Todo: I don't know how
    elif xs.Shape in [CrossSection.SHAPES.RECT_OPEN, CrossSection.SHAPES.RECT_CLOSED]:
        return  # Todo: Rect
    else:
        return shape_generator.swmm_std_cross_sections(xs.Shape, height=VIRTUAL_LENGTH)


def profil_area(inp: SwmmInput, link_label: str):
    cs = get_cross_section_maker(inp, link_label)
    if cs is None:
        return
    xs = inp.XSECTIONS[link_label]

    if xs.Shape == CrossSection.SHAPES.CUSTOM:
        return cs.area_v / VIRTUAL_LENGTH ** 2 * xs.Geom1
    elif xs.Shape == CrossSection.SHAPES.IRREGULAR:
        return  # Todo: I don't know how
    elif xs.Shape in [CrossSection.SHAPES.RECT_OPEN, CrossSection.SHAPES.RECT_CLOSED]:
        return xs.Geom1 * xs.Geom2
    else:
        return cs.area_v / VIRTUAL_LENGTH ** 2 * xs.Geom1


def velocity(inp: SwmmInput, link_label, flow):
    cs = get_cross_section_maker(inp, link_label)
    if cs is None:
        return
    xs = inp.XSECTIONS[link_label]



from io import SEEK_SET

import pandas as pd

from ._read_bin import BinaryReader
from .input_file import SEC
from .output_file.extract import _FLOW_UNITS

_FILESTAMP = "SWMM5-HOTSTART4"


class SwmmHotstart(BinaryReader):
    """The class that handles all extraction of data from the out file."""

    def __init__(self, filename,  inp):
        """
        read a with SWMM created binary hotstart file

        Args:
            filename (str): path of the hotstart-file
            inp (swmm_api.SwmmInput): inp-file-data
        """
        super().__init__(filename)
        self.fp.seek(0, SEEK_SET)

        # openHotstartFile2
        _file_stamp = self._next(len(_FILESTAMP), 's')
        n_subcatchments = self._next()
        n_landuse = self._next()
        n_nodes = self._next()
        n_links = self._next()
        n_pollutants = self._next()
        i_flow_unit = self._next()
        self.unit = _FLOW_UNITS[i_flow_unit]

        # ---------
        # get inp-file-data informations
        pollutants = list(inp.POLLUTANTS.keys())
        landuses = list(inp.LANDUSES.keys())
        labels_subcatchments = list(inp[SEC.SUBCATCHMENTS].keys())

        kind_nodes = []
        node_labels = []

        for sec in inp._original_section_order:
            if sec in [SEC.JUNCTIONS, SEC.OUTFALLS, SEC.STORAGE, SEC.DIVIDERS] and sec in inp:
                kind_nodes += [sec] * len(inp[sec].keys())
                node_labels += list(inp[sec].keys())

        kind_links = []
        links_labels = []
        for sec in inp._original_section_order:
            if sec in [SEC.CONDUITS, SEC.ORIFICES, SEC.PUMPS, SEC.WEIRS, SEC. OUTLETS] and sec in inp:
                kind_links += [sec] * len(inp[sec].keys())
                links_labels += list(inp[sec].keys())

        # ---------------------------------------------------------------------------
        # Runoff

        subareas = ['imperv_zero', 'imperv', 'perv']
        self.columns_subcatchment = ['label', *[f'depth_{s}' for s in subareas], 'runoff']
        # Infiltration
        self.columns_subcatchment += [f'Infiltration_{i}' for i in range(6)]

        if SEC.GROUNDWATER in inp:
            self.columns_subcatchment += ['theta', 'bottomElev+lowerDepth', 'newFlow', 'maxInfilVol']

        if SEC.SNOWPACKS in inp:
            for s in [subareas]:
                for v in ['depth_snow', 'depth_free_water_snow', 'cold_content',
                                              'antecedent_temperature', 'initial_AWESI']:

                    self.columns_subcatchment.append(f'{s}_{v}')

        if n_pollutants:
            for v in ['runoff', 'ponded']:
                for i_pollutant in range(n_pollutants):
                    self.columns_subcatchment.append(f'{v}_{pollutants[i_pollutant]}')

            for i_landuse in range(n_landuse):
                self.columns_subcatchment += [f'{landuses[i_landuse]}_{pollutants[i_pollutant]}_buildup' for i_pollutant in range(n_pollutants)]
                self.columns_subcatchment += [f'{landuses[i_landuse]}_lastSwept']

        self.subcatchments = []
        for i_sc in range(n_subcatchments):
            subcatchment = [labels_subcatchments[i_sc]]
            # Ponded depths for each sub-area & total runoff (4 elements)
            #   impervious w/o depression storage
            #   impervious w/ depression storage
            #   pervious
            subcatchment += self._next_doubles(4)

            # Infiltration state (max. of 6 elements)
            # immer 6 elemente, aber nur die ersten belegt

            #   HORTON:
            #       tp present time on infiltration curve (sec)
            #       Fe cumulative infiltration (ft)
            #   GREEN_AMPT:
            #       IMD current initial soil moisture deficit
            #       F   current cumulative infiltrated volume (ft)
            #       Fu  current upper zone infiltrated volume (ft)
            #       Sat saturation flag
            #       T   time until start of next rain event (sec)
            #   CURVE_NUMBER:
            #       S   current infiltration capacity (ft)
            #       P   current cumulative precipitation (ft)
            #       F   current cumulative infiltration (ft)
            #       T   current inter-event time (sec)
            #       Se  current event infiltration capacity (ft)
            #       f   previous infiltration rate (ft/sec)

            subcatchment += self._next_doubles(6)

            # Groundwater state (4 elements)
            if SEC.GROUNDWATER in inp:
                if labels_subcatchments[i_sc] in inp[SEC.GROUNDWATER]:
                    # theta
                    # bottomElev + lowerDepth
                    # newFlow
                    # maxInfilVol
                    subcatchment += self._next_doubles(4)
                else:
                    subcatchment += [None] * 4

            # Snowpack state (5 elements for each of 3 snow surfaces)
            if SEC.SNOWPACKS in inp:
                if labels_subcatchments[i_sc] in inp[SEC.SNOWPACKS]:
                    for i_snow_surface in range(3):
                        # depth_snow depth of snow pack (ft)
                        # depth_free_water_snow    depth of free water in snow pack (ft)
                        # cold_content cold content of snow pack
                        # antecedent_temperature   antecedent temperature index (deg F)
                        # initial_AWESI   initial AWESI of linear ADC
                        subcatchment += self._next_doubles(5)
                else:
                    subcatchment += [None] * 3*5

            if n_pollutants:
                # Water quality
                #   Runoff quality
                #   Ponded quality
                #   Buildup and when streets were last swept
                subcatchment += self._next_doubles(2*n_pollutants)

                for i_landuse in range(n_landuse):
                    subcatchment += self._next_doubles(n_pollutants + 1)

            self.subcatchments.append(subcatchment)

        # ---------------------------------------------------------------------------
        # Routing
        self.columns_node = ['label', 'kind', 'depth', 'lateral_flow'] + pollutants
        self.columns_storage = ['label', 'kind', 'depth', 'lateral_flow', 'hydraulic_residence_time'] + pollutants

        self.nodes = []
        self.storages = []

        for i_node in range(n_nodes):
            if kind_nodes[i_node] == SEC.STORAGE:
                # if is storage
                # read 3 Parameters
                #   depth
                #   lateral_flow
                #   hydraulic_residence_time
                # + Pollutants
                self.storages.append((node_labels[i_node], kind_nodes[i_node], *self._next_floats(n_pollutants + 3)))
            else:
                # read 2 Parameters
                #   depth
                #   lateral_flow
                #  + Pollutants
                self.nodes.append((node_labels[i_node], kind_nodes[i_node], *self._next_floats(n_pollutants + 2)))

        self.columns_link = ['label', 'kind', 'flow', 'depth', 'setting'] + pollutants

        self.links = []
        for i_link in range(n_links):
            # read 3 Parameters
            #   flow
            #   depth
            #   setting
            #  + Pollutants
            self.links.append((links_labels[i_link], kind_links[i_link], *self._next_floats(n_pollutants + 3)))

    @property
    def links_frame(self):
        df = pd.DataFrame.from_records(self.links)
        df.columns = self.columns_link
        return df

    @property
    def nodes_frame(self):
        df = pd.DataFrame.from_records(self.nodes)
        df.columns = self.columns_node
        return df

    @property
    def storages_frame(self):
        df = pd.DataFrame.from_records(self.storages)
        df.columns = self.columns_storage
        return df

    @property
    def subcatchments_frame(self):
        df = pd.DataFrame.from_records(self.subcatchments)
        df.columns = self.columns_subcatchment
        return df

    def __repr__(self):
        return f'SwmmHotstart(file="{self.filename}")'



from typing import NamedTuple

from numpy import NaN, isnan
from pandas import DataFrame

from ._identifiers import IDENTIFIERS
from ..helpers import BaseSectionObject
from .._type_converter import to_bool, GIS_FLOAT_FORMAT
from ..section_labels import XSECTIONS, LOSSES, VERTICES


class CrossSection(BaseSectionObject):
    """
    Section: [**XSECTIONS**]

    Purpose:
        Provides cross-section geometric data for conduit and regulator links of the drainage system.

    Formats:
        ::

            Link Shape      Geom1 Geom2             Geom3 Geom4 (Barrels Culvert)
            Link CUSTOM     Geom1       Curve                   (Barrels)
            Link IRREGULAR                    Tsect

    Format-PCSWMM:
        ``Link Shape Geom1 Geom2 Geom3 Geom4 (Barrels Culvert)``

    Remarks:
        The Culvert code number is used only for conduits that act as culverts
        and should be analyzed for inlet control conditions using the FHWA HDS-5 method.

        The ``CUSTOM`` shape is a closed conduit whose width versus height is described by a user-supplied Shape Curve.

        An ``IRREGULAR`` cross-section is used to model an open channel whose geometry is described by a Transect object.

    Attributes:
        Link (str): name of the conduit, orifice, or weir.
        Shape (str): cross-section shape (see Tables D-1 below or 3-1 for available shapes).
        Geom1 (float): full height of the cross-section (ft or m).
        Geom2-4: auxiliary parameters (width, side slopes, etc.) as listed in Table D-1.
        Barrels (int): number of barrels (i.e., number of parallel pipes of equal size, slope, and roughness) associated with a conduit (default is 1).
        Culvert (int): code number from Table A.10 for the conduit’s inlet geometry if it is a culvert subject to possible inlet flow control (leave blank otherwise).
        Curve (str): name of a Shape Curve in the [``CURVES``] section that defines how width varies with depth.
        Tsect (str): name of an entry in the [``TRANSECTS``] section that describes the crosssection geometry of an irregular channel.
    """
    _identifier =IDENTIFIERS.Link
    _section_label = XSECTIONS

    class SHAPES:
        IRREGULAR = 'IRREGULAR'  # TransectCoordinates (Natural Channel)
        CUSTOM = 'CUSTOM'  # Full Height, ShapeCurveCoordinates

        CIRCULAR = 'CIRCULAR'  # Full Height = Diameter
        FORCE_MAIN = 'FORCE_MAIN'  # Full Height = Diameter, Roughness
        FILLED_CIRCULAR = 'FILLED_CIRCULAR'  # Full Height = Diameter, Filled Depth
        RECT_CLOSED = 'RECT_CLOSED'  # Rectangular: Full Height, Top Width
        RECT_OPEN = 'RECT_OPEN'  # Rectangular: Full Height, Top Width
        TRAPEZOIDAL = 'TRAPEZOIDAL'  # Full Height, Base Width, Side Slopes
        TRIANGULAR = 'TRIANGULAR'  # Full Height, Top Width
        HORIZ_ELLIPSE = 'HORIZ_ELLIPSE'  # Full Height, Max. Width
        VERT_ELLIPSE = 'VERT_ELLIPSE'  # Full Height, Max. Width
        ARCH = 'ARCH'  # Size Code or Full Height, Max. Width
        PARABOLIC = 'PARABOLIC'  # Full Height, Top Width
        POWER = 'POWER'  # Full Height, Top Width, Exponent
        RECT_TRIANGULAR = 'RECT_TRIANGULAR'  # Full Height, Top Width, Triangle Height
        RECT_ROUND = 'RECT_ROUND'  # Full Height, Top Width, Bottom Radius
        MODBASKETHANDLE = 'MODBASKETHANDLE'  # Full Height, Bottom Width, Top Radius
        EGG = 'EGG'  # Full Height
        HORSESHOE = 'HORSESHOE'  # Full Height Gothic Full Height
        GOTHIC = 'GOTHIC'  # Full Height
        CATENARY = 'CATENARY'  # Full Height
        SEMIELLIPTICAL = 'SEMIELLIPTICAL'  # Full Height
        BASKETHANDLE = 'BASKETHANDLE'  # Full Height
        SEMICIRCULAR = 'SEMICIRCULAR'  # Full Height

    def __init__(self, Link, Shape, Geom1=0, Geom2=0, Geom3=0, Geom4=0, Barrels=1, Culvert=NaN, Tsect=None, Curve=None):
        # in SWMM C-code function "link_readXsectParams"
        self.Link = str(Link)
        self.Shape = Shape

        self.Geom1 = NaN
        self.Tsect = NaN

        self.Geom2 = NaN
        self.Curve = NaN

        if Shape == self.SHAPES.IRREGULAR:
            if Tsect is None:
                # read inp file
                Tsect = Geom1
            self.Tsect = str(Tsect)  # name of TransectCoordinates
        elif Shape == self.SHAPES.CUSTOM:
            if Curve is None:
                Curve = Geom2
            self.Curve = str(Curve)
            self.Geom1 = float(Geom1)
        else:
            self.Geom1 = float(Geom1)
            self.Geom2 = float(Geom2)

        self.Geom3 = float(Geom3)
        self.Geom4 = float(Geom4)
        self.Barrels = int(Barrels)
        # according to the c code 6 arguments are needed to not raise an error / nonsense, but you have to
        if Barrels != 1 or not isinstance(Barrels, str) and ~isnan(Barrels):
            self.Barrels = int(Barrels)
        self.Culvert = Culvert

    @classmethod
    def Irregular(cls, Link, Tsect):
        """
        ``IRREGULAR`` cross-section is used to model an open channel whose geometry is described by a Transect object.
        """
        return cls(Link, CrossSection.SHAPES.IRREGULAR, Tsect=Tsect)

    @classmethod
    def Custom(cls, Link, Geom1, Curve):
        """
        ``CUSTOM`` shape is a closed conduit whose width versus height is described by a user-supplied Shape Curve.
        """
        return cls(Link, CrossSection.SHAPES.CUSTOM, Geom1=Geom1, Curve=Curve)


class Loss(BaseSectionObject):
    """
    Section: [**LOSSES**]

    Purpose:
        Specifies minor head loss coefficients, flap gates, and seepage rates for conduits.

    Formats:
        ::

            Conduit Kentry Kexit Kavg (Flap Seepage)

    Format-PCSWMM:
        ``Link Inlet Outlet Average FlapGate SeepageRate``

    Formats-SWMM-GUI:
        ``Link  Kentry  Kexit  Kavg  FlapGate  Seepage``

    Remarks:
        Minor losses are only computed for the Dynamic Wave flow routing option (see
        [OPTIONS] section). They are computed as Kv 2 /2g where K = minor loss coefficient, v
        = velocity, and g = acceleration of gravity. Entrance losses are based on the velocity
        at the entrance of the conduit, exit losses on the exit velocity, and average losses on
        the average velocity.

        Only enter data for conduits that actually have minor losses, flap valves, or seepage
        losses.

    Args:
        Link (str): name of conduit ``Conduit``
        Inlet (float): entrance minor head loss coefficient. ``Kentry``
        Outlet (float): exit minor head loss coefficient. ``Kexit``
        Average (float): average minor head loss coefficient across length of conduit. ``Kavg``
        FlapGate (bool): YES if conduit has a flap valve that prevents back flow, NO otherwise. (Default is NO). ``Flap``
        SeepageRate (float): Rate of seepage loss into surrounding soil (in/hr or mm/hr). (Default is 0.) ``Seepage``

    Attributes:
        Link (str): name of conduit ``Conduit``
        Inlet (float): entrance minor head loss coefficient. ``Kentry``
        Outlet (float): exit minor head loss coefficient. ``Kexit``
        Average (float): average minor head loss coefficient across length of conduit. ``Kavg``
        FlapGate (bool): YES if conduit has a flap valve that prevents back flow, NO otherwise. (Default is NO). ``Flap``
        SeepageRate (float): Rate of seepage loss into surrounding soil (in/hr or mm/hr). (Default is 0.) ``Seepage``
    """
    _identifier =IDENTIFIERS.Link
    _section_label = LOSSES

    class TYPES:
        INLET = 'Inlet'
        OUTLET = 'Outlet'
        AVERAGE = 'Average'

    def __init__(self, Link, Inlet=0, Outlet=0, Average=0, FlapGate=False, SeepageRate=0):
        self.Link = str(Link)
        self.Inlet = float(Inlet)
        self.Outlet = float(Outlet)
        self.Average = float(Average)
        self.FlapGate = to_bool(FlapGate)
        self.SeepageRate = float(SeepageRate)


class Vertices(BaseSectionObject):
    """
    Section: [**VERTICES**]

    Purpose:
        Assigns X,Y coordinates to interior vertex points of curved drainage system links.

    Format:
        ::

            Link Xcoord Ycoord

    Remarks:
        Straight-line links have no interior vertices and therefore are not listed in this section.

        Include a separate line for each interior vertex of the link, ordered from the inlet node to the outlet
        node.

    Args:
        Link (str): name of link.
        vertices (list[list[float, float]]): vertices relative to origin in lower left of map.

            - Xcoord: horizontal coordinate
            - Ycoord: vertical coordinate

    Attributes:
        Link (str): name of link.
        vertices (list[list[float, float]]): vertices relative to origin in lower left of map.

            - Xcoord: horizontal coordinate
            - Ycoord: vertical coordinate
    """
    _identifier =IDENTIFIERS.Link
    _table_inp_export = False
    _section_label = VERTICES

    def __init__(self, Link,  vertices):
        self.Link = str(Link)
        self.vertices = vertices

    @classmethod
    def _convert_lines(cls, multi_line_args):
        vertices = list()
        last = None

        for line in multi_line_args:
            Link, x, y = line
            x = float(x)
            y = float(y)
            if Link == last:
                vertices.append((x, y))
            else:
                if last is not None:
                    yield cls(last, vertices)
                last = Link
                vertices = [(x, y)]
        # last
        if last is not None:
            yield cls(last, vertices)

    def to_inp_line(self):
        return '\n'.join([f'{self.Link} {x:{GIS_FLOAT_FORMAT}} {y:{GIS_FLOAT_FORMAT}}' for x, y in self.vertices])

    @property
    def frame(self):
        """convert Vertices object to a data-frame

        for debugging purposes

        Returns:
            pandas.DataFrame: section as table
        """
        return DataFrame.from_records(self.vertices, columns=['x', 'y'])



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

from pandas import read_parquet, MultiIndex, Series
import numpy as np

"""
extension to pandas parquet reader and writer

because MultiIndex and integers are not supported as index/columns


index.freq & index/column name are not implemented by parquet
2019-01-12
"""


def _check_name(filename):
    """
    check if name has a common parquet file-extension
    Args:
        filename (str): old filename with or without extension

    Returns:
        str: new filename with extension
    """
    if not (filename.endswith('.parq') or filename.endswith('.parquet')):
        filename += '.parq'
    return filename


def _multiindex_to_index(multiindex):
    """

    Args:
        multiindex (pandas.MultiIndex): old index with multiple level

    Returns:
        pandas.Index: new index with one levels
    """
    if isinstance(multiindex, MultiIndex):
        # compact_name = '/'.join(str(c) for c in multiindex.names)
        multiindex = ['/'.join(str(c) for c in col).strip() for col in multiindex.values]
        # multiindex.name = compact_name
    return multiindex


def write(data, filename, compression='brotli'):
    """
    write data to parquet

    Args:
        data (pandas.DataFrame):
        filename (str): path to resulting file
        compression (str):
    """
    filename = _check_name(filename)
    if isinstance(data, Series):
        df = data.to_frame()
    else:
        df = data.copy()

    df.index = _multiindex_to_index(df.index)
    df.columns = _multiindex_to_index(df.columns)

    df.to_parquet(filename, compression=compression)


def _index_to_multiindex(index):
    """

    Args:
        index (pandas.Index): old index with one level

    Returns:
        pandas.MultiIndex: new index with multiple levels
    """
    if (index.dtype == np.object) and index.str.contains('/').all():
        # old_name = index.name
        index = MultiIndex.from_tuples([col.split('/') for col in index])
        # if isinstance(old_name, str):
        #     new_names = old_name.split('/')
        #     index.names = new_names
    return index


def read(filename):
    """
    read parquet file

    Args:
        filename (str): path to parquet file

    Returns:
        pandas.DataFrame: data
    """
    filename = _check_name(filename)
    df = read_parquet(filename)
    df.columns = _index_to_multiindex(df.columns)
    df.index = _index_to_multiindex(df.index)
    return df



__version__ = '0.2.0.18.2'
from .input_file import read_inp_file, SwmmInput
from .report_file import read_rpt_file, SwmmReport
from .output_file import read_out_file, SwmmOutput, out2frame
from .run import swmm5_run
from .hotstart import SwmmHotstart



from itertools import product

import pandas as pd

from ..inp import SwmmInput
from ..section_labels import TAGS, SUBCATCHMENTS
from .collection import nodes_dict, links_dict
from ..sections import Tag


def _get_tags_frame(inp, part=None):
    if TAGS in inp:
        df_tags = inp[TAGS].frame
        if part in df_tags.index.levels[0]:
            return inp[TAGS].frame.xs(part, axis=0, level=0)
    return pd.Series(name='tags', dtype=str)


def get_node_tags(inp):
    """
    get node tags as pandas.Series

    Args:
        inp (SwmmInput): inp data

    Returns:
        pandas.Series: node tags
    """
    return _get_tags_frame(inp, Tag.TYPES.Node)


def get_link_tags(inp):
    """
    get node link as pandas.Series

    Args:
        inp (SwmmInput): inp data

    Returns:
        pandas.Series: link tags
    """
    return _get_tags_frame(inp, Tag.TYPES.Link)


def get_subcatchment_tags(inp):
    """
    get subcatchment tags as pandas.Series

    Args:
        inp (SwmmInput): inp data

    Returns:
        pandas.Series: subcatchment tags
    """
    return _get_tags_frame(inp, Tag.TYPES.Subcatch)


def filter_tags(inp_tags: SwmmInput, inp_objects: SwmmInput = None):
    """
    get tags of one inp data for objects of another inp data and create new section

    Args:
        inp_tags (SwmmInput): inp data where all tags are
        inp_objects (SwmmInput): inp data of the needed objects

    Returns:
        InpSection[str, Tag] | dict[str, Tag]:
    """
    if inp_objects is None:
        inp_objects = inp_tags

    nodes = nodes_dict(inp_objects)
    keys = list(product([Tag.TYPES.Node], list(nodes.keys())))

    links = links_dict(inp_objects)
    keys += list(product([Tag.TYPES.Link], list(links.keys())))

    if SUBCATCHMENTS in inp_objects:
        keys += list(product([Tag.TYPES.Subcatch], list(inp_objects.SUBCATCHMENTS.keys())))

    return inp_tags.TAGS.slice_section(keys)


def delete_tag_group(inp, part):
    """
    delete all tags of one group

    Args:
        inp (SwmmInput):
        part (str): label of the group i.e.: :attr:`Tag.TYPE.Node` (or the strings: 'Node', 'Subcatch', 'Link')
    """
    if TAGS in inp:
        for key in list(inp[TAGS].keys()):
            if inp[TAGS][key].kind == part:
                del inp[TAGS][key]
